import ast
import csv
import json
import os
import subprocess
from pathlib import Path

import typer
from tqdm import tqdm

from relik.common.log import get_logger
from relik.inference.data.splitters.blank_sentence_splitter import BlankSentenceSplitter
from relik.inference.data.splitters.spacy_sentence_splitter import SpacySentenceSplitter
from relik.inference.data.splitters.window_based_splitter import WindowSentenceSplitter
from relik.inference.data.tokenizers.spacy_tokenizer import SpacyTokenizer
from relik.inference.data.window.manager import WindowManager
from relik.retriever.indexers.document import DocumentStore

logger = get_logger(__name__)

app = typer.Typer(no_args_is_help=True, pretty_exceptions_show_locals=False)


@app.command(context_settings=dict(ignore_unknown_options=True, allow_extra_args=True))
def create_windows(
    input_file: str,
    output_file: str,
    window_size: int = 32,
    window_stride: int = 16,
    title_mapping: str = None,
    relation_mapping: str = None,
    language: str = "en",
    tokenizer_device: str = "cpu",
    is_split_into_words: bool = False,
    labels_are_tokens: bool = False,
    write_batch_size: int = 10_000,
    overwrite: bool = False,
):
    """
    Create windows from input documents and save them to an output file.

    Args:
        input_file (str):
            Path to the input file containing the documents.
        output_file (str):
            Path to the output file to save the windowized data.
        window_size (int, optional):
            Size of the window. Defaults to 32.
        window_stride (int, optional):
            Stride of the window. Defaults to 16.
        title_mapping (str, optional):
            Path to a JSON file containing a mapping of labels. Defaults to None.
        language (str, optional):
            Language of the documents. Defaults to "en".
        tokenizer_device (str, optional):
            Device to use for tokenization. Defaults to "cpu".
        is_split_into_words (bool, optional):
            Whether the documents are already split into words. Defaults to False.
        write_batch_size (int, optional):
            Number of windows to process and write at a time. Defaults to 10_000.

    Returns:
        None
    """

    def _process_batch(
        data,
        window_manager,
        window_size,
        window_stride,
        is_split_into_words,
        title_mapping,
        relation_mapping,
        labels_are_tokens,
    ):
        # build a doc_id to doc mapping
        doc_id_to_doc = {int(document["doc_id"]): document for document in data}
        if is_split_into_words:
            text_field = "doc_words"
        else:
            text_field = "doc_text"

        windowized_data, _, _ = window_manager.create_windows(
            [document[text_field] for document in data],
            window_size,
            window_stride,
            is_split_into_words=is_split_into_words,
            doc_ids=[int(document["doc_id"]) for document in data],
            # doc_topic=doc_topic,
        )

        for window in windowized_data:
            # try:
            # we need to add the labels
            doc_level_labels = doc_id_to_doc[window._d["doc_id"]][
                "doc_span_annotations"
            ]
            doc_text = doc_id_to_doc[window._d["doc_id"]][text_field]
            # if we have a title mapping, we need to map the labels to the
            # new titles
            if title_mapping is not None:
                # compute the missing labels
                # missing_labels |= set(title_mapping.keys()) - set(
                #     [label for _, _, label in doc_level_labels]
                # )
                doc_level_labels = [
                    [start, end, title_mapping.get(label, label)]
                    for start, end, label in doc_level_labels
                ]

            if relation_mapping is not None:
                doc_level_labels = [
                    [start, end, relation_mapping.get(label, label)]
                    for start, end, label in doc_level_labels
                ]
            # these are the labels for the whole document, we need add them to the correct window
            # for window in windowized_document:
            window_level_labels = []
            window_level_labels_but_for_tokens = []
            if not labels_are_tokens:
                for doc_level_label in doc_level_labels:
                    start_char, end_char, label_text = doc_level_label
                    while (
                        len(doc_text) > end_char - 1
                        and end_char - 1 >= 0
                        and doc_text[end_char - 1] == " "
                    ):
                        end_char -= 1
                    while (
                        len(doc_text) > start_char
                        and start_char >= 0
                        and doc_text[start_char] == " "
                    ):
                        start_char += 1
                    if start_char >= window.offset and end_char <= window.offset + len(
                        window.text
                    ):
                        if (
                            start_char > end_char
                            or start_char < window.offset
                            or end_char > window.offset + len(window.text)
                        ):
                            print(
                                f"Error in window {window._d['window_id']}, start: {start_char}, end: {end_char}, window start: {window.offset}, window end: {window.offset + len(window.text)}"
                            )
                            continue
                        window_level_labels.append([start_char, end_char, label_text])
                window._d["window_labels"] = window_level_labels
                # now we need to map the labels to the tokens
                for label in window_level_labels:
                    start_char, end_char, label_text = label
                    start_token = None
                    end_token = None
                    for token_id, (start, end) in enumerate(
                        zip(
                            window._d["token2char_start"].values(),
                            window._d["token2char_end"].values(),
                        )
                    ):
                        if start_char == start:
                            start_token = token_id
                        if end_char == end:
                            end_token = token_id + 1
                        # if we found the tokens, we can break
                        if start_token is not None and end_token is not None:
                            break
                    if start_token is None or end_token is None:
                        # raise ValueError(
                        #     f"Could not find token for label: {label} in window: {window}"
                        # )
                        # warnings.warn(
                        #     f"Could not find token for label: {label} in window: {window}. We will snap the label to the word boundaries."
                        # )
                        start_token = None
                        end_token = None
                        for token_id, (start, end) in enumerate(
                            zip(
                                window._d["token2char_start"].values(),
                                window._d["token2char_end"].values(),
                            )
                        ):
                            if start_char >= start and start_char <= end:
                                start_token = token_id
                            if end_char >= start and end_char <= end:
                                end_token = token_id + 1
                            if start_token is not None and end_token is not None:
                                break
                    if start_token is None or end_token is None:
                        print(
                            f"Error in window {window._d['window_id']}, start: {start_char}, end: {end_char}, window start: {window.offset}, window end: {window.offset + len(window.text)}"
                        )
                        continue
                    window_level_labels_but_for_tokens.append(
                        [start_token, end_token, label_text]
                    )
                window._d["window_labels_tokens"] = window_level_labels_but_for_tokens
            else:
                # if the text is split into words, we need to map the labels to the tokens
                for doc_level_label in doc_level_labels:
                    start_token, end_token, label_text = doc_level_label
                    window_token_start = window._d["char2token_start"][
                        str(window.offset)
                    ]
                    window_token_end = window._d["char2token_end"][
                        str(window.offset + len(window.text))
                    ]
                    if (
                        start_token >= window_token_start
                        and end_token <= window_token_end
                    ):
                        window_level_labels_but_for_tokens.append(doc_level_label)
                window._d["window_labels_tokens"] = window_level_labels_but_for_tokens
                # if the text is split into words, we need to map the labels to the characters
                for label in window_level_labels_but_for_tokens:
                    start_token, end_token, label_text = label
                    start_char = window._d["token2char_start"][str(start_token)]
                    end_char = window._d["token2char_end"][str(end_token - 1)]
                    window_level_labels.append([start_char, end_char, label_text])
                window._d["window_labels"] = window_level_labels

            if "doc_triplet_annotations" in doc_id_to_doc[window._d["doc_id"]]:
                doc_level_triplet_labels = doc_id_to_doc[window._d["doc_id"]][
                    "doc_triplet_annotations"
                ]
                if title_mapping is not None:
                    doc_level_triplet_labels = [
                        {
                            "subject": [
                                triplet["subject"][0],
                                triplet["subject"][1],
                                title_mapping.get(
                                    triplet["subject"][2], triplet["subject"][2]
                                ),
                            ],
                            "relation": title_mapping.get(
                                triplet["relation"], triplet["relation"]
                            ),
                            "object": [
                                triplet["object"][0],
                                triplet["object"][1],
                                title_mapping.get(
                                    triplet["object"][2], triplet["object"][2]
                                ),
                            ],
                        }
                        for triplet in doc_level_triplet_labels
                    ]

                window_level_triplet_labels = []
                window_level_triplet_labels_but_for_tokens = []
                if not labels_are_tokens:
                    for doc_level_triplet_label in doc_level_triplet_labels:
                        subject_start_char, subject_end_char, subject_label_text = (
                            doc_level_triplet_label["subject"]
                        )
                        while (
                            doc_text[subject_end_char - 1] == " "
                        ):  # remove trailing spaces
                            subject_end_char -= 1
                        while doc_text[subject_start_char] == " ":
                            subject_start_char += 1
                        relation_label_text = doc_level_triplet_label["relation"]
                        object_start_char, object_end_char, object_label_text = (
                            doc_level_triplet_label["object"]
                        )
                        while (
                            doc_text[object_end_char - 1] == " "
                        ):  # remove trailing spaces
                            object_end_char -= 1
                        while doc_text[object_start_char] == " ":
                            object_start_char += 1
                        if (
                            subject_start_char >= window.offset
                            and subject_end_char <= window.offset + len(window.text)
                            and object_start_char >= window.offset
                            and object_end_char <= window.offset + len(window.text)
                        ):
                            window_level_triplet_labels.append(
                                {
                                    "subject": [
                                        subject_start_char,
                                        subject_end_char,
                                        subject_label_text,
                                    ],
                                    "relation": relation_label_text,
                                    "object": [
                                        object_start_char,
                                        object_end_char,
                                        object_label_text,
                                    ],
                                }
                            )
                    window._d["window_triplet_labels"] = window_level_triplet_labels
                    # now we need to map the labels to the tokens
                    for triplet in window_level_triplet_labels:
                        subject_start_char, subject_end_char, subject_label_text = (
                            triplet["subject"]
                        )
                        relation_label_text = triplet["relation"]
                        object_start_char, object_end_char, object_label_text = triplet[
                            "object"
                        ]
                        subject_start_token = None
                        subject_end_token = None
                        object_start_token = None
                        object_end_token = None
                        for token_id, (start, end) in enumerate(
                            zip(
                                window._d["token2char_start"].values(),
                                window._d["token2char_end"].values(),
                            )
                        ):
                            if subject_start_char == start:
                                subject_start_token = token_id
                            if subject_end_char == end:
                                subject_end_token = token_id + 1
                            if object_start_char == start:
                                object_start_token = token_id
                            if object_end_char == end:
                                object_end_token = token_id + 1
                        if (
                            subject_start_token is None
                            or subject_end_token is None
                            or object_start_token is None
                            or object_end_token is None
                        ):
                            # raise ValueError(
                            #     f"Could not find token for triplet: {triplet} in window: {window}"
                            # )
                            # find entity in window_labels using character offsets, use same index to find it in window_labels_tokens
                            if subject_start_token is None or subject_end_token is None:
                                subject_index = None
                                for idx, label in enumerate(window._d["window_labels"]):
                                    if (
                                        label[0] == subject_start_char
                                        and label[1] == subject_end_char
                                    ):
                                        subject_index = idx
                                        break
                                if subject_index is None:
                                    print(
                                        f"Error in window {window._d['window_id']}, start: {subject_start_char}, end: {subject_end_char}, window start: {window.offset}, window end: {window.offset + len(window.text)}"
                                    )
                                    continue
                                subject_start_token, subject_end_token, _ = window._d[
                                    "window_labels_tokens"
                                ][subject_index]
                            if object_start_token is None or object_end_token is None:
                                object_index = None
                                for idx, label in enumerate(window._d["window_labels"]):
                                    if (
                                        label[0] == object_start_char
                                        and label[1] == object_end_char
                                    ):
                                        object_index = idx
                                        break
                                if object_index is None:
                                    print(
                                        f"Error in window {window._d['window_id']}, start: {object_start_char}, end: {object_end_char}, window start: {window.offset}, window end: {window.offset + len(window.text)}"
                                    )
                                    continue
                                object_start_token, object_end_token, _ = window._d[
                                    "window_labels_tokens"
                                ][object_index]

                        window_level_triplet_labels_but_for_tokens.append(
                            {
                                "subject": [
                                    subject_start_token,
                                    subject_end_token,
                                    subject_label_text,
                                ],
                                "relation": relation_label_text,
                                "object": [
                                    object_start_token,
                                    object_end_token,
                                    object_label_text,
                                ],
                            }
                        )
                    window._d["window_triplet_labels_tokens"] = (
                        window_level_triplet_labels_but_for_tokens
                    )
                else:
                    # if the text is split into words, we need to map the labels to the tokens
                    for doc_level_triplet_label in doc_level_triplet_labels:
                        subject_start_token, subject_end_token, subject_label_text = (
                            doc_level_triplet_label["subject"]
                        )
                        relation_label_text = doc_level_triplet_label["relation"]
                        object_start_token, object_end_token, object_label_text = (
                            doc_level_triplet_label["object"]
                        )
                        window_token_start = window._d["char2token_start"][
                            str(window.offset)
                        ]
                        window_token_end = window._d["char2token_end"][
                            str(window.offset + len(window.text))
                        ]
                        if (
                            subject_start_token >= window_token_start
                            and subject_end_token <= window_token_end
                            and object_end_token >= window_token_start
                            and object_end_token <= window_token_end
                        ):
                            window_level_triplet_labels_but_for_tokens.append(
                                doc_level_triplet_label
                            )
                    window._d["window_triplet_labels_tokens"] = (
                        window_level_triplet_labels_but_for_tokens
                    )
                    # if the text is split into words, we need to map the labels to the characters
                    for triplet in window_level_triplet_labels_but_for_tokens:
                        subject_start_token, subject_end_token, subject_label_text = (
                            triplet["subject"]
                        )
                        relation_label_text = triplet["relation"]
                        object_start_token, object_end_token, object_label_text = (
                            triplet["object"]
                        )
                        subject_start_char = window._d["token2char_start"][
                            str(subject_start_token)
                        ]
                        subject_end_char = window._d["token2char_end"][
                            str(subject_end_token - 1)
                        ]
                        object_start_char = window._d["token2char_start"][
                            str(object_start_token)
                        ]
                        object_end_char = window._d["token2char_end"][
                            str(object_end_token - 1)
                        ]
                        window_level_triplet_labels.append(
                            {
                                "subject": [
                                    subject_start_char,
                                    subject_end_char,
                                    subject_label_text,
                                ],
                                "relation": relation_label_text,
                                "object": [
                                    object_start_char,
                                    object_end_char,
                                    object_label_text,
                                ],
                            }
                        )
                    window._d["window_triplet_labels"] = window_level_triplet_labels

        # except Exception as e:
        #     logger.error(
        #         f"Error processing document {window._d['doc_id']} window {window._d['window_id']}: {e}"
        #     )

        return windowized_data

    if not Path(input_file).exists():
        raise FileNotFoundError(f"Input file {input_file} not found.")

    # windowization stuff
    tokenizer = SpacyTokenizer(language=language, use_gpu=tokenizer_device == "cuda")
    if window_size == "none":
        sentence_splitter = BlankSentenceSplitter()
    elif window_size == "sentence":
        sentence_splitter = SpacySentenceSplitter()
    else:
        sentence_splitter = WindowSentenceSplitter(
            window_size=window_size, window_stride=window_stride
        )
    window_manager = WindowManager(tokenizer, sentence_splitter)

    if title_mapping is not None:
        with open(title_mapping) as f:
            title_mapping = json.load(f)

    if relation_mapping is not None:
        with open(relation_mapping) as f:
            relation_mapping = json.load(f)

    output_file_path = Path(output_file)

    # check if file exists
    continue_from_id = None
    if output_file_path.exists():
        if not overwrite:
            # we should not overwrite the file
            # open last line of the file using tail command
            try:
                last_line = subprocess.check_output(
                    f"tail -n 1 {output_file}", shell=True
                )
                continue_from_id = json.loads(last_line)["doc_id"]
            except Exception as e:
                raise ValueError(
                    f"Could not read the last line of the output file {output_file}: {e}"
                )

            logger.info(
                f"Output file {output_file} already exists. Continuing from doc id {continue_from_id}"
            )
        else:
            logger.info(
                f"{output_file} already exists but `--overwrite` flag is set, overwriting the file."
            )
    else:
        output_file_path.parent.mkdir(parents=True, exist_ok=True)
        logger.info(f"Saving windowized data to {output_file}")

    logger.info(f"Loading data from {input_file}")
    batched_data = []

    # get number of lines in the file
    # run bash command to get the number of lines in the file
    try:
        total_lines = int(
            subprocess.check_output(
                f"wc -l {input_file} | awk '{{print $1}}'", shell=True
            )
        )
    except Exception as e:
        logger.error(f"Error getting number of lines in the file: {e}")
        total_lines = None

    progress_bar = tqdm(total=total_lines)
    write_mode = "a" if continue_from_id is not None else "w"
    with open(input_file) as f_in, open(output_file, write_mode) as f_out:
        for line in f_in:
            if continue_from_id is not None:
                # we need to skip until we reach the last written line
                current_id = json.loads(line)["doc_id"]
                if current_id != continue_from_id:
                    progress_bar.update(1)
                    continue
                else:
                    continue_from_id = None
            batched_data.append(json.loads(line))
            if len(batched_data) == write_batch_size:
                windowized_data = _process_batch(
                    batched_data,
                    window_manager,
                    window_size,
                    window_stride,
                    is_split_into_words,
                    title_mapping,
                    relation_mapping,
                    labels_are_tokens,
                )
                for wd in windowized_data:
                    f_out.write(wd.to_jsons() + "\n")
                progress_bar.update(len(batched_data))
                batched_data = []

        if len(batched_data) > 0:
            windowized_data = _process_batch(
                batched_data,
                window_manager,
                window_size,
                window_stride,
                is_split_into_words,
                title_mapping,
                relation_mapping,
                labels_are_tokens,
            )
            for wd in windowized_data:
                f_out.write(wd.to_jsons() + "\n")
            progress_bar.update(len(batched_data))
            batched_data = []


@app.command(context_settings=dict(ignore_unknown_options=True, allow_extra_args=True))
def convert_to_dpr(
    input_path: str,
    output_path: str,
    documents_path: str,
    title_map: str = None,
    label_type: str = "span",
):
    if label_type not in ["span", "triplet"]:
        raise ValueError(
            f"Invalid label type: {label_type}. Supported types are: `span`, `triplet`."
        )
    documents = {}
    output_path = Path(output_path)
    output_path.parent.mkdir(parents=True, exist_ok=True)

    # read entities definitions
    logger.info(f"Loading documents from {documents_path}")
    # with open(documents_path, "r") as f:
    #     for line in f:
    #         line_data = json.loads(line)
    #         title = line_data["text"].strip()
    #         definition = line_data["metadata"]["definition"].strip()
    #         documents[title] = definition

    # infer document type
    document_file_type = Path(documents_path).suffix[1:]
    if document_file_type == "jsonl":
        documents = DocumentStore.from_file(documents_path)
    elif document_file_type == "csv":
        documents = DocumentStore.from_tsv(
            documents_path, delimiter=",", quoting=csv.QUOTE_NONE, ingore_case=True
        )
    elif document_file_type == "tsv":
        documents = DocumentStore.from_tsv(
            documents_path, delimiter="\t", quoting=csv.QUOTE_NONE, ingore_case=True
        )
    else:
        raise ValueError(
            f"Unknown document file type: {document_file_type}. Supported types are: jsonl, csv, and tsv."
        )

    if title_map is not None:
        with open(title_map, "r") as f:
            title_map = json.load(f)
    else:
        title_map = {}

    # store dpr data
    dpr = []
    # lower case titles
    title_to_lower_map = {doc.text.lower(): doc.text for doc in documents}
    # store missing entities
    missing = set()
    # Read input file
    with open(input_path, "r") as f, open(output_path, "w") as f_out:
        for line in tqdm(f, desc="Processing data"):
            sentence = json.loads(line)
            # for sentence in aida_data:
            question = (
                sentence["doc_text"] if "doc_text" in sentence else sentence["text"]
            )
            positive_pssgs = []
            if label_type == "triplet":
                for idx, triplet in enumerate(sentence["window_triplet_labels"]):
                    relation = triplet["relation"]
                    if not relation:
                        continue
                    # relation = relation.strip().lower()
                    relation = relation.strip()
                    # relation = title_to_lower_map.get(relation, relation)
                    if relation in documents:
                        # def_text = documents[relation]
                        doc = documents.get_document_from_text(relation)
                        doc.metadata["passage_id"] = (
                            f"{sentence['doc_id']}_{sentence['offset']}_{idx}"
                        )
                        positive_pssgs.append(doc.to_dict())
                    else:
                        missing.add(relation)
                        print(f"Relation {relation} not found in definitions")
            else:
                for idx, entity in enumerate(sentence["window_labels"]):
                    entity = entity[2]

                    if not entity:
                        continue

                    # entity = entity.strip().lower().replace("_", " ")
                    entity = entity.strip()

                    if entity == "--NME--":
                        continue

                    # if title_map and entity in title_to_lower_map:
                    # entity = title_to_lower_map.get(entity, entity)
                    entity = title_map.get(entity, entity)
                    if entity in documents:
                        doc = documents.get_document_from_text(entity)
                        # doc.text = mapped_entity
                        doc.metadata["passage_id"] = (
                            f"{sentence['doc_id']}_{sentence['offset']}_{idx}"
                        )
                        positive_pssgs.append(doc.to_dict())
                    else:
                        missing.add(entity)
                        print(f"Entity {entity} not found in definitions")

            if len(positive_pssgs) == 0:
                continue

            dpr_sentence = {
                "id": f"{sentence['doc_id']}_{sentence['offset']}",
                "doc_topic": sentence["doc_topic"],
                "question": question,
                "positive_ctxs": positive_pssgs,
                "negative_ctxs": "",
                "hard_negative_ctxs": "",
            }
            f_out.write(json.dumps(dpr_sentence) + "\n")

        for e in missing:
            print(e)
        print(f"Number of missing entities: {len(missing)}")

    return dpr


if __name__ == "__main__":
    app()

# import ast
# import csv
# import json
# import os
# import subprocess
# from pathlib import Path

# import typer
# from tqdm import tqdm

# from relik.common.log import get_logger
# from relik.inference.data.splitters.blank_sentence_splitter import BlankSentenceSplitter
# from relik.inference.data.splitters.spacy_sentence_splitter import SpacySentenceSplitter
# from relik.inference.data.splitters.window_based_splitter import WindowSentenceSplitter
# from relik.inference.data.tokenizers.spacy_tokenizer import SpacyTokenizer
# from relik.inference.data.window.manager import WindowManager
# from relik.retriever.indexers.document import DocumentStore

# logger = get_logger(__name__)

# app = typer.Typer(no_args_is_help=True, pretty_exceptions_show_locals=False)


# @app.command(context_settings=dict(ignore_unknown_options=True, allow_extra_args=True))
# def create_windows(
#     input_file: str,
#     output_file: str,
#     window_size: int = typer.Option(
#         32, parser=lambda x: None if x == "None" else int(x)
#     ),
#     window_stride: int = typer.Option(
#         16, parser=lambda x: None if x == "None" else int(x)
#     ),
#     # window_size: int | str = 32,
#     # window_stride: int = 16,
#     title_mapping: str = None,
#     relation_mapping: str = None,
#     language: str = "en",
#     tokenizer_device: str = "cpu",
#     is_split_into_words: bool = False,
#     labels_are_tokens: bool = False,
#     write_batch_size: int = 10_000,
#     overwrite: bool = False,
# ):
#     """
#     Create windows from input documents and save them to an output file.

#     Args:
#         input_file (str):
#             Path to the input file containing the documents.
#         output_file (str):
#             Path to the output file to save the windowized data.
#         window_size (int, optional):
#             Size of the window. Defaults to 32.
#         window_stride (int, optional):
#             Stride of the window. Defaults to 16.
#         title_mapping (str, optional):
#             Path to a JSON file containing a mapping of labels. Defaults to None.
#         language (str, optional):
#             Language of the documents. Defaults to "en".
#         tokenizer_device (str, optional):
#             Device to use for tokenization. Defaults to "cpu".
#         is_split_into_words (bool, optional):
#             Whether the documents are already split into words. Defaults to False.
#         write_batch_size (int, optional):
#             Number of windows to process and write at a time. Defaults to 10_000.

#     Returns:
#         None
#     """

#     def _process_batch(
#         data,
#         window_manager,
#         window_size,
#         window_stride,
#         is_split_into_words,
#         title_mapping,
#         relation_mapping,
#         labels_are_tokens,
#     ):
#         # build a doc_id to doc mapping
#         doc_id_to_doc = {int(document["doc_id"]): document for document in data}
#         if is_split_into_words:
#             text_field = "doc_words"
#         else:
#             text_field = "doc_text"

#         windowized_data, _, _ = window_manager.create_windows(
#             [document[text_field] for document in data],
#             window_size,
#             window_stride,
#             is_split_into_words=is_split_into_words,
#             doc_ids=[int(document["doc_id"]) for document in data],
#             # doc_topic=doc_topic,
#         )

#         for window in windowized_data:
#             # try:
#             # we need to add the labels
#             doc_level_labels = doc_id_to_doc[window._d["doc_id"]][
#                 "doc_span_annotations"
#             ]
#             # if we have a title mapping, we need to map the labels to the
#             # new titles
#             if title_mapping is not None:
#                 # compute the missing labels
#                 # missing_labels |= set(title_mapping.keys()) - set(
#                 #     [label for _, _, label in doc_level_labels]
#                 # )
#                 doc_level_labels = [
#                     [start, end, title_mapping.get(label, label)]
#                     for start, end, label in doc_level_labels
#                 ]

#             if relation_mapping is not None:
#                 doc_level_labels = [
#                     [start, end, relation_mapping.get(label, label)]
#                     for start, end, label in doc_level_labels
#                 ]
#             # these are the labels for the whole document, we need add them to the correct window
#             # for window in windowized_document:
#             window_level_labels = []
#             window_level_labels_but_for_tokens = []
#             if not labels_are_tokens:
#                 for doc_level_label in doc_level_labels:
#                     start_char, end_char, label_text = doc_level_label
#                     if start_char >= window.offset and end_char <= window.offset + len(
#                         window.text
#                     ):
#                         window_level_labels.append(doc_level_label)
#                 window._d["window_labels"] = window_level_labels
#                 # now we need to map the labels to the tokens
#                 for label in window_level_labels:
#                     start_char, end_char, label_text = label
#                     start_token = None
#                     end_token = None
#                     for token_id, (start, end) in enumerate(
#                         zip(
#                             window._d["token2char_start"].values(),
#                             window._d["token2char_end"].values(),
#                         )
#                     ):
#                         if start_char == start:
#                             start_token = token_id
#                         if end_char == end:
#                             end_token = token_id + 1
#                         # if we found the tokens, we can break
#                         if start_token is not None and end_token is not None:
#                             break
#                     if start_token is None or end_token is None:
#                         # raise ValueError(
#                         #     f"Could not find token for label: {label} in window: {window}"
#                         # )
#                         # warnings.warn(
#                         #     f"Could not find token for label: {label} in window: {window}. We will snap the label to the word boundaries."
#                         # )
#                         start_token = None
#                         end_token = None
#                         for token_id, (start, end) in enumerate(
#                             zip(
#                                 window._d["token2char_start"].values(),
#                                 window._d["token2char_end"].values(),
#                             )
#                         ):
#                             if start_char >= start and start_char < end:
#                                 start_token = token_id
#                             if end_char > start and end_char <= end:
#                                 end_token = token_id + 1
#                             if start_token is not None and end_token is not None:
#                                 break
#                     window_level_labels_but_for_tokens.append(
#                         [start_token, end_token, label_text]
#                     )
#                 window._d["window_labels_tokens"] = window_level_labels_but_for_tokens
#             else:
#                 # if the text is split into words, we need to map the labels to the tokens
#                 for doc_level_label in doc_level_labels:
#                     start_token, end_token, label_text = doc_level_label
#                     window_token_start = window._d["char2token_start"][
#                         str(window.offset)
#                     ]
#                     window_token_end = window._d["char2token_end"][
#                         str(window.offset + len(window.text))
#                     ]
#                     if (
#                         start_token >= window_token_start
#                         and end_token <= window_token_end + 1
#                     ):
#                         window_level_labels_but_for_tokens.append(doc_level_label)
#                 window._d["window_labels_tokens"] = window_level_labels_but_for_tokens
#                 # if the text is split into words, we need to map the labels to the characters
#                 for label in window_level_labels_but_for_tokens:
#                     start_token, end_token, label_text = label
#                     start_char = window._d["token2char_start"][str(start_token)]
#                     end_char = window._d["token2char_end"][str(end_token - 1)]
#                     window_level_labels.append([start_char, end_char, label_text])
#                 window._d["window_labels"] = window_level_labels

#             if "doc_triplet_annotations" in doc_id_to_doc[window._d["doc_id"]]:
#                 doc_level_triplet_labels = doc_id_to_doc[window._d["doc_id"]][
#                     "doc_triplet_annotations"
#                 ]
#                 if title_mapping is not None:
#                     doc_level_triplet_labels = [
#                         {
#                             "subject": [
#                                 triplet["subject"][0],
#                                 triplet["subject"][1],
#                                 title_mapping.get(
#                                     triplet["subject"][2], triplet["subject"][2]
#                                 ),
#                             ],
#                             "relation": title_mapping.get(
#                                 triplet["relation"], triplet["relation"]
#                             ),
#                             "object": [
#                                 triplet["object"][0],
#                                 triplet["object"][1],
#                                 title_mapping.get(
#                                     triplet["object"][2], triplet["object"][2]
#                                 ),
#                             ],
#                         }
#                         for triplet in doc_level_triplet_labels
#                     ]

#                 window_level_triplet_labels = []
#                 window_level_triplet_labels_but_for_tokens = []

#                 if not labels_are_tokens:
#                     for doc_level_triplet_label in doc_level_triplet_labels:
#                         subject_start_char, subject_end_char, subject_label_text = (
#                             doc_level_triplet_label["subject"]
#                         )
#                         relation_label_text = doc_level_triplet_label["relation"]
#                         object_start_char, object_end_char, object_label_text = (
#                             doc_level_triplet_label["object"]
#                         )
#                         if (
#                             subject_start_char >= window.offset
#                             and subject_end_char <= window.offset + len(window.text)
#                             and object_start_char >= window.offset
#                             and object_end_char <= window.offset + len(window.text)
#                         ):
#                             window_level_triplet_labels.append(doc_level_triplet_label)
#                     window._d["window_triplet_labels"] = window_level_triplet_labels
#                     # now we need to map the labels to the tokens
#                     for triplet in window_level_triplet_labels:
#                         subject_start_char, subject_end_char, subject_label_text = (
#                             triplet["subject"]
#                         )
#                         relation_label_text = triplet["relation"]
#                         object_start_char, object_end_char, object_label_text = triplet[
#                             "object"
#                         ]
#                         subject_start_token = None
#                         subject_end_token = None
#                         object_start_token = None
#                         object_end_token = None
#                         for token_id, (start, end) in enumerate(
#                             zip(
#                                 window._d["token2char_start"].values(),
#                                 window._d["token2char_end"].values(),
#                             )
#                         ):
#                             if subject_start_char == start:
#                                 subject_start_token = token_id
#                             if subject_end_char == end:
#                                 subject_end_token = token_id + 1
#                             if object_start_char == start:
#                                 object_start_token = token_id
#                             if object_end_char == end:
#                                 object_end_token = token_id + 1
#                         if (
#                             subject_start_token is None
#                             or subject_end_token is None
#                             or object_start_token is None
#                             or object_end_token is None
#                         ):
#                             # raise ValueError(
#                             #     f"Could not find token for triplet: {triplet} in window: {window}"
#                             # )
#                             # find entity in window_labels using character offsets, use same index to find it in window_labels_tokens
#                             if subject_start_token is None or subject_end_token is None:
#                                 subject_index = None
#                                 for idx, label in enumerate(window._d["window_labels"]):
#                                     if (
#                                         label[0] == subject_start_char
#                                         and label[1] == subject_end_char
#                                     ):
#                                         subject_index = idx
#                                         break
#                                 subject_start_token, subject_end_token, _ = window._d[
#                                     "window_labels_tokens"
#                                 ][subject_index]
#                             if object_start_token is None or object_end_token is None:
#                                 object_index = None
#                                 for idx, label in enumerate(window._d["window_labels"]):
#                                     if (
#                                         label[0] == object_start_char
#                                         and label[1] == object_end_char
#                                     ):
#                                         object_index = idx
#                                         break
#                                 object_start_token, object_end_token, _ = window._d[
#                                     "window_labels_tokens"
#                                 ][object_index]

#                         window_level_triplet_labels_but_for_tokens.append(
#                             {
#                                 "subject": [
#                                     subject_start_token,
#                                     subject_end_token,
#                                     subject_label_text,
#                                 ],
#                                 "relation": relation_label_text,
#                                 "object": [
#                                     object_start_token,
#                                     object_end_token,
#                                     object_label_text,
#                                 ],
#                             }
#                         )
#                     window._d["window_triplet_labels_tokens"] = (
#                         window_level_triplet_labels_but_for_tokens
#                     )
#                 else:
#                     # if the text is split into words, we need to map the labels to the tokens
#                     for doc_level_triplet_label in doc_level_triplet_labels:
#                         subject_start_token, subject_end_token, subject_label_text = (
#                             doc_level_triplet_label["subject"]
#                         )
#                         relation_label_text = doc_level_triplet_label["relation"]
#                         object_start_token, object_end_token, object_label_text = (
#                             doc_level_triplet_label["object"]
#                         )
#                         window_token_start = window._d["char2token_start"][
#                             str(window.offset)
#                         ]
#                         window_token_end = window._d["char2token_end"][
#                             str(window.offset + len(window.text))
#                         ]
#                         if (
#                             subject_start_token >= window_token_start
#                             and subject_end_token <= window_token_end + 1
#                             and object_end_token >= window_token_start
#                             and object_end_token <= window_token_end + 1
#                         ):
#                             window_level_triplet_labels_but_for_tokens.append(
#                                 doc_level_triplet_label
#                             )
#                     window._d["window_triplet_labels_tokens"] = (
#                         window_level_triplet_labels_but_for_tokens
#                     )
#                     # if the text is split into words, we need to map the labels to the characters
#                     for triplet in window_level_triplet_labels_but_for_tokens:
#                         subject_start_token, subject_end_token, subject_label_text = (
#                             triplet["subject"]
#                         )
#                         relation_label_text = triplet["relation"]
#                         object_start_token, object_end_token, object_label_text = (
#                             triplet["object"]
#                         )
#                         subject_start_char = window._d["token2char_start"][
#                             str(subject_start_token)
#                         ]
#                         subject_end_char = window._d["token2char_end"][
#                             str(subject_end_token - 1)
#                         ]
#                         object_start_char = window._d["token2char_start"][
#                             str(object_start_token)
#                         ]
#                         object_end_char = window._d["token2char_end"][
#                             str(object_end_token - 1)
#                         ]
#                         window_level_triplet_labels.append(
#                             {
#                                 "subject": [
#                                     subject_start_char,
#                                     subject_end_char,
#                                     subject_label_text,
#                                 ],
#                                 "relation": relation_label_text,
#                                 "object": [
#                                     object_start_char,
#                                     object_end_char,
#                                     object_label_text,
#                                 ],
#                             }
#                         )
#                     window._d["window_triplet_labels"] = window_level_triplet_labels

#         # except Exception as e:
#         #     logger.error(
#         #         f"Error processing document {window._d['doc_id']} window {window._d['window_id']}: {e}"
#         #     )

#         return windowized_data

#     if not Path(input_file).exists():
#         raise FileNotFoundError(f"Input file {input_file} not found.")

#     # windowization stuff
#     tokenizer = SpacyTokenizer(language=language, use_gpu=tokenizer_device == "cuda")
#     if window_size == "none" or window_size is None:
#         sentence_splitter = BlankSentenceSplitter()
#     elif window_size == "sentence":
#         sentence_splitter = SpacySentenceSplitter()
#     else:
#         sentence_splitter = WindowSentenceSplitter(
#             window_size=window_size, window_stride=window_stride
#         )
#     window_manager = WindowManager(tokenizer, sentence_splitter)

#     if title_mapping is not None:
#         with open(title_mapping) as f:
#             title_mapping = json.load(f)

#     if relation_mapping is not None:
#         with open(relation_mapping) as f:
#             relation_mapping = json.load(f)

#     output_file_path = Path(output_file)

#     # check if file exists
#     continue_from_id = None
#     if output_file_path.exists():
#         if not overwrite:
#             # we should not overwrite the file
#             # open last line of the file using tail command
#             try:
#                 last_line = subprocess.check_output(
#                     f"tail -n 1 {output_file}", shell=True
#                 )
#                 continue_from_id = json.loads(last_line)["doc_id"]
#             except Exception as e:
#                 raise ValueError(
#                     f"Could not read the last line of the output file {output_file}: {e}"
#                 )

#             logger.info(
#                 f"Output file {output_file} already exists. Continuing from doc id {continue_from_id}"
#             )
#         else:
#             logger.info(
#                 f"{output_file} already exists but `--overwrite` flag is set, overwriting the file."
#             )
#     else:
#         output_file_path.parent.mkdir(parents=True, exist_ok=True)
#         logger.info(f"Saving windowized data to {output_file}")

#     logger.info(f"Loading data from {input_file}")
#     batched_data = []

#     # get number of lines in the file
#     # run bash command to get the number of lines in the file
#     try:
#         total_lines = int(
#             subprocess.check_output(
#                 f"wc -l {input_file} | awk '{{print $1}}'", shell=True
#             )
#         )
#     except Exception as e:
#         logger.error(f"Error getting number of lines in the file: {e}")
#         total_lines = None

#     progress_bar = tqdm(total=total_lines)
#     write_mode = "a" if continue_from_id is not None else "w"
#     with open(input_file) as f_in, open(output_file, write_mode) as f_out:
#         for line in f_in:
#             if continue_from_id is not None:
#                 # we need to skip until we reach the last written line
#                 current_id = json.loads(line)["doc_id"]
#                 if current_id != continue_from_id:
#                     progress_bar.update(1)
#                     continue
#                 else:
#                     continue_from_id = None
#             batched_data.append(json.loads(line))
#             if len(batched_data) == write_batch_size:
#                 windowized_data = _process_batch(
#                     batched_data,
#                     window_manager,
#                     window_size,
#                     window_stride,
#                     is_split_into_words,
#                     title_mapping,
#                     relation_mapping,
#                     labels_are_tokens,
#                 )
#                 for wd in windowized_data:
#                     f_out.write(wd.to_jsons() + "\n")
#                 progress_bar.update(len(batched_data))
#                 batched_data = []

#         if len(batched_data) > 0:
#             windowized_data = _process_batch(
#                 batched_data,
#                 window_manager,
#                 window_size,
#                 window_stride,
#                 is_split_into_words,
#                 title_mapping,
#                 relation_mapping,
#                 labels_are_tokens,
#             )
#             for wd in windowized_data:
#                 f_out.write(wd.to_jsons() + "\n")
#             progress_bar.update(len(batched_data))
#             batched_data = []


# @app.command(context_settings=dict(ignore_unknown_options=True, allow_extra_args=True))
# def convert_to_dpr(
#     input_path: str,
#     output_path: str,
#     documents_path: str,
#     title_map: str = None,
#     label_type: str = "span",
# ):
#     if label_type not in ["span", "triplet"]:
#         raise ValueError(
#             f"Invalid label type: {label_type}. Supported types are: `span`, `triplet`."
#         )
#     documents = {}
#     output_path = Path(output_path)
#     output_path.parent.mkdir(parents=True, exist_ok=True)

#     # read entities definitions
#     logger.info(f"Loading documents from {documents_path}")
#     # with open(documents_path, "r") as f:
#     #     for line in f:
#     #         line_data = json.loads(line)
#     #         title = line_data["text"].strip()
#     #         definition = line_data["metadata"]["definition"].strip()
#     #         documents[title] = definition

#     # infer document type
#     document_file_type = Path(documents_path).suffix[1:]
#     if document_file_type == "jsonl":
#         documents = DocumentStore.from_file(documents_path)
#     elif document_file_type == "csv":
#         documents = DocumentStore.from_tsv(
#             documents_path, delimiter=",", quoting=csv.QUOTE_NONE, ingore_case=True
#         )
#     elif document_file_type == "tsv":
#         documents = DocumentStore.from_tsv(
#             documents_path, delimiter="\t", quoting=csv.QUOTE_NONE, ingore_case=True
#         )
#     else:
#         raise ValueError(
#             f"Unknown document file type: {document_file_type}. Supported types are: jsonl, csv, and tsv."
#         )

#     if title_map is not None:
#         with open(title_map, "r") as f:
#             title_map = json.load(f)
#     else:
#         title_map = {}

#     # store dpr data
#     dpr = []
#     # lower case titles
#     # title_to_lower_map = {doc.text.lower(): doc.text for doc in documents}
#     # store missing entities
#     missing = set()
#     # Read input file
#     with open(input_path, "r") as f, open(output_path, "w") as f_out:
#         for line in tqdm(f, desc="Processing data"):
#             sentence = json.loads(line)
#             # for sentence in aida_data:
#             question = sentence["text"]
#             positive_pssgs = []

#             if label_type == "triplet":
#                 for idx, triplet in enumerate(sentence["window_triplet_labels"]):
#                     relation = triplet["relation"]
#                     if not relation:
#                         continue
#                     # relation = relation.strip().lower()
#                     relation = relation.strip()
#                     # relation = title_to_lower_map.get(relation, relation)
#                     if relation in documents:
#                         # def_text = documents[relation]
#                         doc = documents.get_document_from_text(relation)
#                         doc.metadata["passage_id"] = (
#                             f"{sentence['doc_id']}_{sentence['offset']}_{idx}"
#                         )
#                         positive_pssgs.append(doc.to_dict())
#                     else:
#                         missing.add(relation)
#                         print(f"Relation {relation} not found in definitions")
#             else:
#                 for idx, entity in enumerate(sentence["window_labels"]):
#                     entity = entity[2]

#                     if not entity:
#                         continue

#                     # entity = entity.strip().lower().replace("_", " ")
#                     entity = entity.strip()

#                     if entity == "--NME--":
#                         continue

#                     # if title_map and entity in title_to_lower_map:
#                     # entity = title_to_lower_map.get(entity, entity)
#                     entity = title_map.get(entity, entity)
#                     if entity in documents:
#                         doc = documents.get_document_from_text(entity)
#                         # doc.text = mapped_entity
#                         doc.metadata["passage_id"] = (
#                             f"{sentence['doc_id']}_{sentence['offset']}_{idx}"
#                         )
#                         positive_pssgs.append(doc.to_dict())
#                     else:
#                         missing.add(entity)
#                         print(f"Entity {entity} not found in definitions")

#             if len(positive_pssgs) == 0:
#                 continue

#             dpr_sentence = {
#                 "id": f"{sentence['doc_id']}_{sentence['offset']}",
#                 "doc_topic": sentence["doc_topic"],
#                 "question": question,
#                 "positive_ctxs": positive_pssgs,
#                 "negative_ctxs": "",
#                 "hard_negative_ctxs": "",
#             }
#             f_out.write(json.dumps(dpr_sentence) + "\n")

#         for e in missing:
#             print(e)
#         print(f"Number of missing entities: {len(missing)}")

#     return dpr


# if __name__ == "__main__":
#     app()



from __future__ import annotations

from collections import UserDict
from typing import Any, Union

import torch
from lightning.fabric.utilities import move_data_to_device

from relik.common.log import get_logger

logger = get_logger(__name__)


class ModelInputs(UserDict):
    """Model input dictionary wrapper."""

    def __getattr__(self, item: str):
        try:
            return self.data[item]
        except KeyError:
            raise AttributeError(f"`ModelInputs` has no attribute `{item}`")

    def __getitem__(self, item: str) -> Any:
        return self.data[item]

    def __getstate__(self):
        return {"data": self.data}

    def __setstate__(self, state):
        if "data" in state:
            self.data = state["data"]

    def keys(self):
        """A set-like object providing a view on D's keys."""
        return self.data.keys()

    def values(self):
        """An object providing a view on D's values."""
        return self.data.values()

    def items(self):
        """A set-like object providing a view on D's items."""
        return self.data.items()

    def to(self, device: Union[str, torch.device]) -> ModelInputs:
        """
        Send all tensors values to device.
        Args:
            device (`str` or `torch.device`): The device to put the tensors on.
        Returns:
            :class:`tokenizers.ModelInputs`: The same instance of :class:`~tokenizers.ModelInputs`
            after modification.
        """
        self.data = move_data_to_device(self.data, device)
        return self



import ast
import json
import os
import subprocess
from pathlib import Path
from typing import Annotated, Optional, Union

import typer
from tqdm import tqdm

from relik.cli import data, reader, retriever
from relik.common.log import get_logger
from relik.common.utils import batch_generator
from relik.inference.annotator import Relik
from relik.inference.data.splitters.blank_sentence_splitter import BlankSentenceSplitter
from relik.inference.data.splitters.spacy_sentence_splitter import SpacySentenceSplitter
from relik.inference.data.splitters.window_based_splitter import WindowSentenceSplitter
from relik.inference.data.tokenizers.spacy_tokenizer import SpacyTokenizer
from relik.inference.data.window.manager import WindowManager
from relik.inference.serve.backend.fastapi_be import main as serve_fastapi
from relik.inference.serve.frontend.gradio_fe import main as serve_gradio

logger = get_logger(__name__)

app = typer.Typer(no_args_is_help=True, pretty_exceptions_show_locals=False)
app.add_typer(data.app, name="data")
app.add_typer(reader.app, name="reader")
app.add_typer(retriever.app, name="retriever")


@app.command(context_settings=dict(ignore_unknown_options=True, allow_extra_args=True))
def inference(
    model_name_or_path: str,
    input_path: str,
    output_path: str,
    batch_size: int = 8,
    num_workers: int = 4,
    device: str = "cuda",
    precision: str = "fp16",
    top_k: int = 100,
    io_batch_size: int = 1_000,
    annotation_type: str = "char",
    progress_bar: bool = True,
    model_kwargs: str = None,
    inference_kwargs: str = None,
):
    """
    Perform inference on raw text files using a pre-trained pipeline.

    Args:
        model_name_or_path (str):
            The model name or path to the model directory.
        input_path (str):
            The path to the input file as txt. Each line in the file should contain a single text sample.
        output_path (str):
            The path to the output file as jsonl. Each line in the file will contain the model's prediction.
        batch_size (int, optional):
            The batch size for inference. Defaults to 8.
        num_workers (int, optional):
            The number of workers for parallel processing. Defaults to 4.
        device (str, optional):
            The device to use for inference (e.g., "cuda", "cpu"). Defaults to "cuda".
        precision (str, optional):
            The precision mode for inference (e.g., "fp16", "fp32"). Defaults to "fp16".
        top_k (int, optional):
            The number of top predictions of the retriever to consider. Defaults to 100.
        window_size (int, optional):
            The window size for sliding window annotation. Defaults to None.
        window_stride (int, optional):
            The stride size for sliding window annotation. Defaults to None.
        annotation_type (str, optional):
            The type of annotation to use (e.g., "CHAR", "WORD"). Defaults to "CHAR".
        progress_bar (bool, optional):
            Whether to display a progress bar during inference. Defaults to True.
        model_kwargs (dict, optional):
            Additional keyword arguments for the model. Defaults to None.
        inference_kwargs (dict, optional):
            Additional keyword arguments for inference. Defaults to None.

    """
    if model_kwargs is None:
        model_kwargs = {}
    else:
        model_kwargs = ast.literal_eval(model_kwargs)

    model_kwargs.update(dict(device=device, precision=precision))

    if inference_kwargs is None:
        inference_kwargs = {}
    else:
        inference_kwargs = ast.literal_eval(inference_kwargs)

    inference_kwargs.update(
        dict(
            num_workers=num_workers,
            top_k=top_k,
            annotation_type=annotation_type,
            progress_bar=progress_bar,
        )
    )
    if "retriever_batch_size" not in inference_kwargs:
        inference_kwargs["retriever_batch_size"] = batch_size
    if "reader_batch_size" not in inference_kwargs:
        inference_kwargs["reader_batch_size"] = batch_size

    # create folder if it doesn't exist
    input_path = Path(input_path)
    output_path = Path(output_path)

    if input_path.is_dir():
        input_files = list(input_path.glob("*.txt"))
    else:
        input_files = [input_path]
    if output_path.is_dir():
        output_path.mkdir(parents=True, exist_ok=True)
    else:
        output_path.parent.mkdir(parents=True, exist_ok=True)

    relik = Relik.from_pretrained(model_name_or_path, **model_kwargs)

    for i, path in enumerate(input_files):
        logger.info("Starting annotation for %s", path)
        if output_path.is_dir():
            output_path = output_path / f"{path.stem}.jsonl"
            out_context = open(output_path, "w")
        else:
            # if we are writing to a single file, we need to append
            # but only if we are doing inference on multiple files
            # and if we are at the second file
            out_context = open(output_path, "w" if i == 0 else "a")
        with open(path, "r") as f_in, out_context as f_out:
            for batch in tqdm(batch_generator(f_in, io_batch_size)):
                # clean batch
                batch = [line.strip() for line in batch if line.strip()]
                predictions = relik(batch, **inference_kwargs)
                for prediction in predictions:
                    prdiction_dict = prediction.to_dict()
                    prediction_dict = {
                        "text": prdiction_dict["text"],
                        "spans": prdiction_dict["spans"],
                    }
                    f_out.write(json.dumps(prediction_dict) + "\n")


@app.command(context_settings=dict(ignore_unknown_options=True, allow_extra_args=True))
def serve(
    relik_pretrained: Annotated[
        str,
        typer.Argument(help="The device to use for relik (e.g., 'cuda', 'cpu')."),
    ],
    device: str = "cpu",
    retriever_device: str = None,
    document_index_device: str = None,
    reader_device: str = None,
    precision: int = 32,
    retriever_precision: int = None,
    document_index_precision: int = None,
    reader_precision: int = None,
    annotation_type: str = "char",
    host: str = "0.0.0.0",
    port: int = 8000,
    frontend: bool = False,
):
    serve_fastapi(
        relik_pretrained=relik_pretrained,
        device=device,
        retriever_device=retriever_device,
        document_index_device=document_index_device,
        reader_device=reader_device,
        precision=precision,
        retriever_precision=retriever_precision,
        document_index_precision=document_index_precision,
        reader_precision=reader_precision,
        annotation_type=annotation_type,
        host=host,
        port=port,
        frontend=frontend,
    )


if __name__ == "__main__":
    app()



import logging
from typing import (
    Any,
    Callable,
    Dict,
    Generator,
    Iterator,
    List,
    NamedTuple,
    Optional,
    Tuple,
    Union,
)

import numpy as np
import torch
import tqdm
from torch.utils.data import IterableDataset
from transformers import AutoTokenizer, PreTrainedTokenizer

from relik.reader.data.relik_reader_data_utils import (
    add_noise_to_value,
    batchify,
    batchify_matrices,
    batchify_tensor,
    chunks,
    flatten,
)
from relik.reader.data.relik_reader_sample import (
    RelikReaderSample,
    load_relik_reader_samples,
)
from relik.reader.utils.special_symbols import NME_SYMBOL

from torch.utils.data import get_worker_info
import torch.distributed as dist

logger = logging.getLogger(__name__)


class TokenizationOutput(NamedTuple):
    input_ids: torch.Tensor
    attention_mask: torch.Tensor
    token_type_ids: torch.Tensor
    prediction_mask: torch.Tensor
    special_symbols_mask: torch.Tensor
    special_symbols_mask_entities: torch.Tensor


class RelikREDataset(IterableDataset):
    def __init__(
        self,
        dataset_path: Optional[str],
        materialize_samples: bool,
        transformer_model: Union[str, PreTrainedTokenizer],
        special_symbols_re: List[str],
        special_symbols: Optional[List[str]] = [],
        shuffle_candidates: Optional[Union[bool, float]] = False,
        flip_candidates: Optional[Union[bool, float]] = False,
        for_inference: bool = False,
        noise_param: float = 0.1,
        sorting_fields: Optional[str] = None,
        tokens_per_batch: int = 2048,
        batch_size: int = None,
        max_batch_size: int = 128,
        section_size: int = 500_000,
        prebatch: bool = True,
        add_gold_candidates: bool = True,
        use_nme: bool = False,
        min_length: int = -1,
        max_length: int = 2048,
        max_triplets: int = 50,
        max_spans: int = 100,
        model_max_length: int = 2048,
        skip_empty_training_samples: bool = True,
        drop_last: bool = False,
        samples: Optional[Iterator[RelikReaderSample]] = None,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.dataset_path = dataset_path
        self.materialize_samples = materialize_samples
        self.samples: Optional[List[RelikReaderSample]] = samples
        if self.materialize_samples and self.samples is None:
            self.samples = list()

        self.special_symbols = special_symbols_re
        self.special_symbols_types = special_symbols

        if isinstance(transformer_model, str):
            self.tokenizer = self._build_tokenizer(
                transformer_model, self.special_symbols + self.special_symbols_types
            )
        else:
            self.tokenizer = transformer_model

        self.shuffle_candidates = shuffle_candidates
        self.flip_candidates = flip_candidates
        self.for_inference = for_inference
        self.noise_param = noise_param
        self.batching_fields = ["input_ids"]
        self.sorting_fields = (
            sorting_fields if sorting_fields is not None else self.batching_fields
        )
        self.add_gold_candidates = add_gold_candidates
        self.use_nme = use_nme
        self.min_length = min_length
        self.max_length = max_length
        self.model_max_length = (
            model_max_length
            if model_max_length < self.tokenizer.model_max_length
            else self.tokenizer.model_max_length
        )
        self.transformer_model = transformer_model
        self.skip_empty_training_samples = skip_empty_training_samples
        self.drop_last = drop_last

        self.tokens_per_batch = tokens_per_batch
        self.batch_size = batch_size
        self.max_batch_size = max_batch_size
        self.max_triplets = max_triplets
        self.max_spans = max_spans
        self.section_size = section_size
        self.prebatch = prebatch

    def _build_tokenizer(self, transformer_model: str, special_symbols: List[str]):
        return AutoTokenizer.from_pretrained(
            transformer_model,
            additional_special_tokens=[ss for ss in special_symbols],
            add_prefix_space=True,
        )

    @staticmethod
    def get_special_symbols_re(num_entities: int, use_nme: bool = False) -> List[str]:
        if use_nme:
            return [NME_SYMBOL] + [f"[R-{i}]" for i in range(num_entities)]
        else:
            return [f"[R-{i}]" for i in range(num_entities)]

    @staticmethod
    def get_special_symbols(num_entities: int) -> List[str]:
        return [NME_SYMBOL] + [f"[E-{i}]" for i in range(num_entities)]

    @property
    def fields_batcher(self) -> Dict[str, Union[None, Callable[[list], Any]]]:
        fields_batchers = {
            "input_ids": lambda x: batchify(
                x, padding_value=self.tokenizer.pad_token_id
            ),
            "attention_mask": lambda x: batchify(x, padding_value=0),
            "token_type_ids": lambda x: batchify(x, padding_value=0),
            "prediction_mask": lambda x: batchify(x, padding_value=1),
            "global_attention": lambda x: batchify(x, padding_value=0),
            "token2word": None,
            "sample": None,
            "special_symbols_mask": lambda x: batchify(x, padding_value=False),
            "special_symbols_mask_entities": lambda x: batchify(x, padding_value=False),
            "start_labels": lambda x: batchify(x, padding_value=-100),
            "end_labels": lambda x: batchify_matrices(x, padding_value=-100),
            "disambiguation_labels": lambda x: batchify(x, padding_value=-100),
            "relation_labels": lambda x: batchify_tensor(x, padding_value=-100),
            "predictable_candidates": None,
        }
        if (
            isinstance(self.transformer_model, str)
            and "roberta" in self.transformer_model
        ) or (
            isinstance(self.transformer_model, PreTrainedTokenizer)
            and "roberta" in self.transformer_model.config.model_type
        ):
            del fields_batchers["token_type_ids"]

        return fields_batchers

    def _build_input_ids(
        self, sentence_input_ids: List[int], candidates_input_ids: List[List[int]]
    ) -> List[int]:
        return (
            [self.tokenizer.cls_token_id]
            + sentence_input_ids
            + [self.tokenizer.sep_token_id]
            + flatten(candidates_input_ids)
            + [self.tokenizer.sep_token_id]
        )

    def _build_input(self, text: List[str], candidates: List[List[str]]) -> List[int]:
        return (
            text
            + [self.tokenizer.sep_token]
            + flatten(candidates)
            + [self.tokenizer.sep_token]
        )

    def _build_tokenizer_essentials(
        self, input_ids, original_sequence, ents=0
    ) -> TokenizationOutput:
        input_ids = torch.tensor(input_ids, dtype=torch.long)
        attention_mask = torch.ones_like(input_ids)

        if len(self.special_symbols_types) > 0:
            # special symbols mask
            special_symbols_mask = input_ids >= self.tokenizer.vocab_size
            # select only the first N true values where N is len(entities_definitions)
            special_symbols_mask_entities = special_symbols_mask.clone()
            special_symbols_mask_entities[
                special_symbols_mask_entities.cumsum(0) > ents
            ] = False
            token_type_ids = (torch.cumsum(special_symbols_mask, dim=0) > 0).long()
            special_symbols_mask = special_symbols_mask ^ special_symbols_mask_entities
        else:
            special_symbols_mask = input_ids >= self.tokenizer.vocab_size
            special_symbols_mask_entities = special_symbols_mask.clone()
            token_type_ids = (torch.cumsum(special_symbols_mask, dim=0) > 0).long()

        prediction_mask = token_type_ids.roll(shifts=-1, dims=0)
        prediction_mask[-1] = 1
        prediction_mask[0] = 1

        assert len(prediction_mask) == len(input_ids)

        return TokenizationOutput(
            input_ids,
            attention_mask,
            token_type_ids,
            prediction_mask,
            special_symbols_mask,
            special_symbols_mask_entities,
        )

    @staticmethod
    def _subindex(lst, target_values, dims):
        for i, sublist in enumerate(lst):
            match = all(sublist[dim] == target_values[dim] for dim in dims)
            if match:
                return i

    def _build_labels(
        self,
        sample,
        tokenization_output: TokenizationOutput,
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        len_input_ids = len(tokenization_output.input_ids)
        start_labels = [0] * len_input_ids
        end_labels = []
        end_labels_tensor = [0] * len_input_ids

        # Pre-process entities to handle edge cases and duplicates.
        sample.window_labels_tokens.sort(key=lambda x: (x[0], x[1]))
        processed_entities = []
        entities_untyped = set()
        for entity in sample.window_labels_tokens:
            entity = self._adjust_entity_boundaries(entity, sample.word2token)
            if entity:
                processed_entities.append(entity)
                entities_untyped.add((entity[0], entity[1]))

        # Initialize disambiguation and relation labels.
        entities_untyped = sorted(entities_untyped)
        num_disambiguation_labels = (
            len(sample.span_candidates) + len(sample.triplet_candidates)
            if self.special_symbols_types
            else len(sample.triplet_candidates)
        )
        disambiguation_labels = torch.zeros(
            (len(entities_untyped), num_disambiguation_labels)
        )

        # Main loop to fill in start and end labels.
        prev_start_bpe = -1
        offset = 0
        for idx, entity in enumerate(processed_entities):
            start_bpe, end_bpe = self._get_bpe_positions(entity, sample.word2token)
            class_index = idx + 1  # Offset by 1 for NONE class

            # Update start labels
            start_labels[start_bpe] = class_index

            # Handle end labels, considering overlapping entities.
            if start_bpe != prev_start_bpe:
                end_labels.append(end_labels_tensor.copy())
                end_labels[-1][:start_bpe] = [-100] * start_bpe
            elif end_labels[-1][end_bpe] != 0:
                # Handle overlapping entities by adjusting the start position.
                offset += 1
                prev_start_bpe = start_bpe
                continue
            end_labels[-1][end_bpe] = class_index

            # Update disambiguation labels if applicable.
            if self.special_symbols_types and entity[2] in sample.span_candidates:
                disambiguation_labels[
                    idx - offset, sample.span_candidates.index(entity[2])
                ] = 1
            prev_start_bpe = start_bpe

        # Handle ignored indices for start and end labels.
        ignored_indices = tokenization_output.prediction_mask == 1
        start_labels_tensor = torch.tensor(start_labels, dtype=torch.long)
        start_labels_tensor[ignored_indices] = -100
        end_labels_tensor = torch.stack([torch.tensor(label) for label in end_labels])
        end_labels_tensor[ignored_indices.repeat(len(end_labels), 1)] = -100

        # Initialize and update relation labels.
        relation_labels = self._init_relation_labels(entities_untyped, sample)

        return (
            start_labels_tensor,
            end_labels_tensor,
            disambiguation_labels,
            relation_labels,
        )

    def _adjust_entity_boundaries(self, entity, word2token):
        """Adjust entity boundaries to ensure they map to valid token positions."""
        # Adjust start position
        while not word2token.get(entity[0], []):
            entity = (entity[0] + 1, entity[1], entity[2])
            if entity[0] >= len(word2token):
                return None
        # Adjust end position
        while not word2token.get(entity[1] - 1, []):
            entity = (entity[0], entity[1] - 1, entity[2])
            if entity[1] <= 0:
                return None
        return entity

    def _get_bpe_positions(self, entity, word2token):
        """Get the start and end positions in BPE tokens for an entity."""
        return word2token[entity[0]][0] + 1, word2token[entity[1] - 1][-1] + 1

    def _init_relation_labels(self, entities_untyped, sample):
        """
        Initialize the tensor for relation labels between entities.
        """
        num_entities = len(entities_untyped)
        num_relations = len(sample.triplet_candidates)
        relation_labels = torch.zeros((num_entities, num_entities, num_relations))

        if sample.window_triplet_labels_tokens is None:
            relation_labels.fill_(-100)
            return relation_labels

        for relation in sample.window_triplet_labels_tokens:
            # Determine the index of this relation type in the triplet_candidates list.
            # If the relation type is not found, use a default index (e.g., for 'unknown' relation types).
            relation_idx = (
                sample.triplet_candidates.index(relation["relation"])
                if relation["relation"] in sample.triplet_candidates
                else -1
            )

            # Find the indices of the subject and object entities within the entities_untyped list.
            subject_idx = next(
                (
                    i
                    for i, entity in enumerate(entities_untyped)
                    if entity[0] == relation["subject"][0]
                    and entity[1] == relation["subject"][1]
                ),
                None,
            )
            object_idx = next(
                (
                    i
                    for i, entity in enumerate(entities_untyped)
                    if entity[0] == relation["object"][0]
                    and entity[1] == relation["object"][1]
                ),
                None,
            )

            # If both the subject and object entity indices are found, mark the relation in the tensor.
            if (
                subject_idx is not None
                and object_idx is not None
                and relation_idx != -1
            ):
                relation_labels[subject_idx, object_idx, relation_idx] = 1

        return relation_labels

    def __iter__(self):
        dataset_iterator = self.dataset_iterator_func()
        i = None
        if self.section_size is not None:
            current_dataset_elements = []
            for i, dataset_elem in enumerate(dataset_iterator, start=1):
                if len(current_dataset_elements) == self.section_size:
                    for batch in self.materialize_batches(current_dataset_elements):
                        yield batch
                    current_dataset_elements = []
                current_dataset_elements.append(dataset_elem)
                if i % 50_000 == 0:
                    logger.info(f"Processed: {i} number of elements")
            if len(current_dataset_elements) != 0:
                for batch in self.materialize_batches(current_dataset_elements):
                    yield batch
            if i is not None:
                logger.debug(f"Dataset finished: {i} number of elements processed")
            else:
                logger.warning("Dataset empty")
        else:
            for batch in self.materialize_batches(dataset_iterator):
                if i is None:
                    i = 0
                i += batch["input_ids"].shape[0]
                yield batch
            if i is not None:
                logger.debug(f"Dataset finished: {i} number of elements processed")
            else:
                logger.warning("Dataset empty")

    def dataset_iterator_func(self):
        data_samples = (
            load_relik_reader_samples(self.dataset_path)
            if self.samples is None
            or (isinstance(self.samples, list) and len(self.samples) == 0)
            else self.samples
        )
        if self.materialize_samples:
            data_acc = []
        # take care of the tqdm nesting
        # for sample in tqdm.tqdm(data_samples, desc="Reading dataset"):

        # Handle DDP
        world_size = 1
        rank = 0
        if dist.is_available() and dist.is_initialized():
            world_size = dist.get_world_size()
            rank = dist.get_rank()

        def partition_data_samples(iterable, world_size, rank):
            def generator():
                for i, x in enumerate(iterable):
                    if i % world_size == rank:
                        yield x

            return generator

        samples_per_rank = partition_data_samples(data_samples, world_size, rank)()

        for sample in samples_per_rank:
            if self.materialize_samples and sample.materialize is not None:
                # tokenization_output = sample.materialize["tokenization_output"]
                materialized = sample.materialize
                del sample.materialize
                yield {
                    "input_ids": materialized["tokenization_output"].input_ids,
                    "attention_mask": materialized[
                        "tokenization_output"
                    ].attention_mask,
                    "token_type_ids": materialized[
                        "tokenization_output"
                    ].token_type_ids,
                    "prediction_mask": materialized[
                        "tokenization_output"
                    ].prediction_mask,
                    "special_symbols_mask": materialized[
                        "tokenization_output"
                    ].special_symbols_mask,
                    "special_symbols_mask_entities": materialized[
                        "tokenization_output"
                    ].special_symbols_mask_entities,
                    "sample": sample,
                    "start_labels": materialized["start_labels"],
                    "end_labels": materialized["end_labels"],
                    "disambiguation_labels": materialized["disambiguation_labels"],
                    "relation_labels": materialized["relation_labels"],
                    "predictable_candidates": materialized["candidates_symbols"],
                }
                sample.materialize = materialized
                data_acc.append(sample)
                continue
            candidates_symbols = self.special_symbols
            candidates_entities_symbols = self.special_symbols_types

            # sample.candidates = sample.candidates[: self.max_candidates]

            if len(self.special_symbols_types) > 0:
                # sample.span_candidates = sample.span_candidates[
                #     : self.max_ent_candidates
                # ]
                # add NME as a possible candidate
                assert sample.span_candidates is not None
                if self.use_nme:
                    sample.span_candidates.insert(0, NME_SYMBOL)
                    # sample.candidates.insert(0, NME_SYMBOL)

            sample.triplet_candidates = sample.triplet_candidates[
                : min(len(candidates_symbols), self.max_triplets)
            ]

            if len(self.special_symbols_types) > 0:
                sample.span_candidates = sample.span_candidates[
                    : min(len(candidates_entities_symbols), self.max_spans)
                ]
            # training time sample mods
            if not self.for_inference:
                # check whether the sample has labels if not skip
                if (
                    sample.window_triplet_labels_tokens is None
                    or len(sample.window_triplet_labels_tokens) == 0
                ) and self.skip_empty_training_samples:
                    # logger.warning(
                    #     "Sample {} has no labels, skipping".format(sample.doc_id)
                    # )
                    continue

                # add gold candidates if missing
                if self.add_gold_candidates and sample.window_triplet_labels_tokens:
                    candidates_set = set(sample.triplet_candidates)
                    candidates_to_add = set()
                    for candidate_title in sample.window_triplet_labels_tokens:
                        if candidate_title["relation"] not in candidates_set:
                            candidates_to_add.add(candidate_title["relation"])
                    if len(candidates_to_add) > 0:
                        # replacing last candidates with the gold ones
                        # this is done in order to preserve the ordering
                        candidates_to_add = list(candidates_to_add)
                        added_gold_candidates = 0
                        gold_candidates_titles_set = set(
                            set(
                                ct["relation"]
                                for ct in sample.window_triplet_labels_tokens
                            )
                        )
                        for i in reversed(range(len(sample.triplet_candidates))):
                            if (
                                sample.triplet_candidates[i]
                                not in gold_candidates_titles_set
                                and sample.triplet_candidates[i] != NME_SYMBOL
                            ):
                                sample.triplet_candidates[i] = candidates_to_add[
                                    added_gold_candidates
                                ]
                                added_gold_candidates += 1
                                if len(candidates_to_add) == added_gold_candidates:
                                    break

                        candidates_still_to_add = (
                            len(candidates_to_add) - added_gold_candidates
                        )
                        while (
                            len(sample.triplet_candidates)
                            <= min(len(candidates_symbols), self.max_triplets)
                            and candidates_still_to_add != 0
                        ):
                            sample.triplet_candidates.append(
                                candidates_to_add[added_gold_candidates]
                            )
                            added_gold_candidates += 1
                            candidates_still_to_add -= 1

                def shuffle_cands(shuffle_candidates, candidates):
                    if (
                        isinstance(shuffle_candidates, bool) and shuffle_candidates
                    ) or (
                        isinstance(shuffle_candidates, float)
                        and np.random.uniform() < shuffle_candidates
                    ):
                        np.random.shuffle(candidates)
                        if NME_SYMBOL in candidates:
                            candidates.remove(NME_SYMBOL)
                            candidates.insert(0, NME_SYMBOL)
                    return candidates

                def flip_cands(flip_candidates, candidates):
                    # flip candidates
                    if (isinstance(flip_candidates, bool) and flip_candidates) or (
                        isinstance(flip_candidates, float)
                        and np.random.uniform() < flip_candidates
                    ):
                        for i in range(len(candidates) - 1):
                            if np.random.uniform() < 0.5:
                                candidates[i], candidates[i + 1] = (
                                    candidates[i + 1],
                                    candidates[i],
                                )
                        if NME_SYMBOL in candidates:
                            candidates.remove(NME_SYMBOL)
                            candidates.insert(0, NME_SYMBOL)
                    return candidates

                if self.shuffle_candidates:
                    sample.triplet_candidates = shuffle_cands(
                        self.shuffle_candidates, sample.triplet_candidates
                    )
                    if len(self.special_symbols_types) > 0:
                        sample.span_candidates = shuffle_cands(
                            self.shuffle_candidates, sample.span_candidates
                        )
                elif self.flip_candidates:
                    sample.triplet_candidates = flip_cands(
                        self.flip_candidates, sample.triplet_candidates
                    )
                    if len(self.special_symbols_types) > 0:
                        sample.span_candidates = flip_cands(
                            self.flip_candidates, sample.span_candidates
                        )

            # candidates encoding
            candidates_symbols = candidates_symbols[: len(sample.triplet_candidates)]

            candidates_encoding = [
                ["{} {}".format(cs, ct)] if ct != NME_SYMBOL else [NME_SYMBOL]
                for cs, ct in zip(candidates_symbols, sample.triplet_candidates)
            ]
            if len(self.special_symbols_types) > 0:
                candidates_entities_symbols = candidates_entities_symbols[
                    : len(sample.span_candidates)
                ]
                candidates_types_encoding = [
                    ["{} {}".format(cs, ct)] if ct != NME_SYMBOL else [NME_SYMBOL]
                    for cs, ct in zip(
                        candidates_entities_symbols, sample.span_candidates
                    )
                ]
                candidates_encoding = (
                    candidates_types_encoding
                    + [[self.tokenizer.sep_token]]
                    + candidates_encoding
                )

            pretoken_input = self._build_input(sample.words, candidates_encoding)
            input_tokenized = self.tokenizer(
                pretoken_input,
                return_offsets_mapping=True,
                add_special_tokens=False,
            )

            window_tokens = input_tokenized.input_ids
            window_tokens = flatten(window_tokens)

            offsets_mapping = [
                [
                    (
                        ss + sample.token2char_start[str(i)],
                        se + sample.token2char_start[str(i)],
                    )
                    for ss, se in input_tokenized.offset_mapping[i]
                ]
                for i in range(len(sample.words))
            ]

            offsets_mapping = flatten(offsets_mapping)

            token2char_start = {str(i): s for i, (s, _) in enumerate(offsets_mapping)}
            token2char_end = {str(i): e for i, (_, e) in enumerate(offsets_mapping)}
            token2word_start = {
                str(i): int(sample._d["char2token_start"][str(s)])
                for i, (s, _) in enumerate(offsets_mapping)
                if str(s) in sample._d["char2token_start"]
            }
            token2word_end = {
                str(i): int(sample._d["char2token_end"][str(e)])
                for i, (_, e) in enumerate(offsets_mapping)
                if str(e) in sample._d["char2token_end"]
            }
            # invert token2word_start and token2word_end
            word2token_start = {str(v): int(k) for k, v in token2word_start.items()}
            word2token_end = {str(v): int(k) for k, v in token2word_end.items()}

            sample._d.update(
                dict(
                    tokens=window_tokens,
                    token2char_start=token2char_start,
                    token2char_end=token2char_end,
                    token2word_start=token2word_start,
                    token2word_end=token2word_end,
                    word2token_start=word2token_start,
                    word2token_end=word2token_end,
                )
            )

            input_subwords = flatten(input_tokenized["input_ids"][: len(sample.words)])
            offsets = input_tokenized["offset_mapping"][: len(sample.words)]
            token2word = []
            word2token = {}
            count = 0
            for i, offset in enumerate(offsets):
                word2token[i] = []
                for token in offset:
                    token2word.append(i)
                    word2token[i].append(count)
                    count += 1

            sample.token2word = token2word
            sample.word2token = word2token
            candidates_encoding_result = input_tokenized["input_ids"][
                len(sample.words) + 1 : -1
            ]

            i = 0
            cum_len = 0
            # drop candidates if the number of input tokens is too long for the model
            if (
                sum(map(len, candidates_encoding_result))
                + len(input_subwords)
                + 20  # + 20 special tokens
                > self.model_max_length
            ):
                if self.for_inference:
                    acceptable_tokens_from_candidates = (
                        self.model_max_length - 20 - len(input_subwords)
                    )
                    current_len = (
                        len(candidates_encoding_result[i])
                        + len(
                            candidates_encoding_result[i + len(sample.span_candidates)]
                        )
                        if len(self.special_symbols_types) > 0
                        else len(candidates_encoding_result[i])
                    )
                    while cum_len + current_len < acceptable_tokens_from_candidates:
                        cum_len += current_len
                        i += 1
                        if len(self.special_symbols_types) == 0 or i + len(
                            sample.span_candidates
                        ) >= len(candidates_encoding_result):
                            current_len = len(candidates_encoding_result[i])
                        else:
                            current_len = len(
                                candidates_encoding_result[i]
                                + candidates_encoding_result[
                                    i + len(sample.span_candidates)
                                ]
                            )

                    assert i > 0

                    candidates_encoding_result = (
                        candidates_encoding_result[:i]
                        + candidates_encoding_result[len(sample.span_candidates) : i]
                    )
                    if len(self.special_symbols_types) > 0:
                        candidates_entities_symbols = candidates_entities_symbols[:i]
                        sample.span_candidates = sample.span_candidates[:i]

                    candidates_symbols = candidates_symbols[:i]
                    sample.triplet_candidates = sample.triplet_candidates[:i]
                else:
                    if len(sample.window_triplet_labels_tokens) == 0:
                        sample.window_triplet_labels_tokens = []
                    gold_candidates_set = set(
                        [wl["relation"] for wl in sample.window_triplet_labels_tokens]
                    )
                    gold_candidates_indices = [
                        i
                        for i, wc in enumerate(sample.triplet_candidates)
                        if wc in gold_candidates_set
                    ]
                    if len(self.special_symbols_types) > 0:
                        gold_candidates_indices = [
                            i + len(sample.span_candidates)
                            for i in gold_candidates_indices
                        ] + [len(sample.span_candidates)]
                        gold_candidates_set_entities = set(
                            [wl[2] for wl in sample.window_labels_tokens]
                        )
                        gold_candidates_indices += [
                            i
                            for i, wc in enumerate(sample.span_candidates)
                            if wc in gold_candidates_set_entities
                        ]
                        # # add entities indices
                        # gold_candidates_indices = gold_candidates_indices + list(
                        #     range(len(sample.span_candidates))
                        # )
                    necessary_taken_tokens = sum(
                        map(
                            len,
                            [
                                candidates_encoding_result[i]
                                for i in gold_candidates_indices
                            ],
                        )
                    )

                    acceptable_tokens_from_candidates = (
                        self.model_max_length
                        - 20
                        - len(input_subwords)
                        - necessary_taken_tokens
                    )
                    if acceptable_tokens_from_candidates <= 0:
                        logger.warning(
                            "Sample {} has no candidates after truncation due to max length".format(
                                sample.doc_id
                            )
                        )
                        continue
                    # assert acceptable_tokens_from_candidates > 0

                    i = 0
                    cum_len = 0
                    while (
                        cum_len + len(candidates_encoding_result[i])
                        < acceptable_tokens_from_candidates
                    ):
                        if i not in gold_candidates_indices:
                            cum_len += len(candidates_encoding_result[i])
                        i += 1

                    new_indices = sorted(
                        list(set(list(range(i)) + gold_candidates_indices))
                    )
                    # np.random.shuffle(new_indices)

                    candidates_encoding_result = [
                        candidates_encoding_result[i] for i in new_indices
                    ]
                    if len(self.special_symbols_types) > 0:
                        sample.triplet_candidates = [
                            sample.triplet_candidates[i - len(sample.span_candidates)]
                            for i in new_indices[len(sample.span_candidates) : -1]
                        ]
                        candidates_symbols = candidates_symbols[
                            : len(sample.triplet_candidates)
                        ]
                    else:
                        candidates_symbols = [
                            candidates_symbols[i] for i in new_indices
                        ]
                        sample.triplet_candidates = [
                            sample.triplet_candidates[i] for i in new_indices
                        ]
                if len(sample.triplet_candidates) == 0:
                    logger.warning(
                        "Sample {} has no candidates after truncation due to max length".format(
                            sample.doc_id
                        )
                    )
                    continue

            # final input_ids build
            input_ids = self._build_input_ids(
                sentence_input_ids=input_subwords,
                candidates_input_ids=candidates_encoding_result,
            )

            # complete input building (e.g. attention / prediction mask)
            tokenization_output = self._build_tokenizer_essentials(
                input_ids,
                input_subwords,
                (
                    min(len(sample.span_candidates), len(self.special_symbols_types))
                    if sample.span_candidates is not None
                    else 0
                ),
            )
            # labels creation
            start_labels, end_labels, disambiguation_labels, relation_labels = (
                None,
                None,
                None,
                None,
            )
            if (
                sample.window_labels_tokens is not None
                and len(sample.window_labels_tokens) > 0
            ):
                (
                    start_labels,
                    end_labels,
                    disambiguation_labels,
                    relation_labels,
                ) = self._build_labels(
                    sample,
                    tokenization_output,
                )
            elif not self.for_inference:
                continue

            if self.materialize_samples:
                sample.materialize = {
                    "tokenization_output": tokenization_output,
                    "start_labels": start_labels,
                    "end_labels": end_labels,
                    "disambiguation_labels": disambiguation_labels,
                    "relation_labels": relation_labels,
                    "candidates_symbols": candidates_symbols,
                }
                data_acc.append(sample)
            yield {
                "input_ids": tokenization_output.input_ids,
                "attention_mask": tokenization_output.attention_mask,
                "token_type_ids": tokenization_output.token_type_ids,
                "prediction_mask": tokenization_output.prediction_mask,
                "special_symbols_mask": tokenization_output.special_symbols_mask,
                "special_symbols_mask_entities": tokenization_output.special_symbols_mask_entities,
                "sample": sample,
                "start_labels": start_labels,
                "end_labels": end_labels,
                "disambiguation_labels": disambiguation_labels,
                "relation_labels": relation_labels,
                "predictable_candidates": candidates_symbols
                + candidates_entities_symbols,
            }
        if self.materialize_samples:
            self.samples = data_acc

    def preshuffle_elements(self, dataset_elements: List):
        # This shuffling is done so that when using the sorting function,
        # if it is deterministic given a collection and its order, we will
        # make the whole operation not deterministic anymore.
        # Basically, the aim is not to build every time the same batches.
        if not self.for_inference:
            dataset_elements = np.random.permutation(dataset_elements)

        sorting_fn = lambda elem: (
            add_noise_to_value(
                sum(len(elem[k]) for k in self.sorting_fields),
                noise_param=self.noise_param,
            )
            if not self.for_inference
            else sum(len(elem[k]) for k in self.sorting_fields)
        )

        dataset_elements = sorted(dataset_elements, key=sorting_fn)

        if self.for_inference:
            return dataset_elements

        ds = list(chunks(dataset_elements, 64))  # todo: modified
        np.random.shuffle(ds)
        return flatten(ds)

    def materialize_batches(
        self, dataset_elements: List[Dict[str, Any]]
    ) -> Generator[Dict[str, Any], None, None]:
        if self.prebatch and self.section_size is not None and not self.for_inference:
            dataset_elements = self.preshuffle_elements(dataset_elements)

        current_batch = []

        # function that creates a batch from the 'current_batch' list
        def output_batch() -> Dict[str, Any]:
            assert (
                len(
                    set([len(elem["predictable_candidates"]) for elem in current_batch])
                )
                == 1
            ), " ".join(
                map(
                    str, [len(elem["predictable_candidates"]) for elem in current_batch]
                )
            )

            batch_dict = dict()

            de_values_by_field = {
                fn: [de[fn] for de in current_batch if fn in de]
                for fn in self.fields_batcher
            }

            # in case you provide fields batchers but in the batch
            # there are no elements for that field
            de_values_by_field = {
                fn: fvs for fn, fvs in de_values_by_field.items() if len(fvs) > 0
            }

            assert len(set([len(v) for v in de_values_by_field.values()]))

            # todo: maybe we should report the user about possible
            #  fields filtering due to "None" instances
            de_values_by_field = {
                fn: fvs
                for fn, fvs in de_values_by_field.items()
                if all([fv is not None for fv in fvs])
            }

            for field_name, field_values in de_values_by_field.items():
                field_batch = (
                    self.fields_batcher[field_name](field_values)
                    if self.fields_batcher[field_name] is not None
                    else field_values
                )

                batch_dict[field_name] = field_batch

            return batch_dict

        max_len_discards, min_len_discards = 0, 0

        should_token_batch = self.batch_size is None

        curr_pred_elements = -1
        for de in dataset_elements:
            if (
                should_token_batch
                and self.max_batch_size != -1
                and len(current_batch) == self.max_batch_size
            ) or (not should_token_batch and len(current_batch) == self.batch_size):
                yield output_batch()
                current_batch = []
                curr_pred_elements = -1

            # todo support max length (and min length) as dicts

            too_long_fields = [
                k
                for k in de
                if self.max_length != -1
                and torch.is_tensor(de[k])
                and len(de[k]) > self.max_length
            ]
            if len(too_long_fields) > 0:
                max_len_discards += 1
                continue

            too_short_fields = [
                k
                for k in de
                if self.min_length != -1
                and torch.is_tensor(de[k])
                and len(de[k]) < self.min_length
            ]
            if len(too_short_fields) > 0:
                min_len_discards += 1
                continue

            if should_token_batch:
                de_len = sum(len(de[k]) for k in self.batching_fields)

                future_max_len = max(
                    de_len,
                    max(
                        [
                            sum(len(bde[k]) for k in self.batching_fields)
                            for bde in current_batch
                        ],
                        default=0,
                    ),
                )

                future_tokens_per_batch = future_max_len * (len(current_batch) + 1)

                num_predictable_candidates = len(de["predictable_candidates"])

                if len(current_batch) > 0 and (
                    future_tokens_per_batch >= self.tokens_per_batch
                    or (
                        num_predictable_candidates != curr_pred_elements
                        and curr_pred_elements != -1
                    )
                ):
                    yield output_batch()
                    current_batch = []

            current_batch.append(de)
            curr_pred_elements = len(de["predictable_candidates"])

        if len(current_batch) != 0 and not self.drop_last:
            yield output_batch()

        if max_len_discards > 0:
            if self.for_inference:
                logger.warning(
                    f"WARNING: Inference mode is True but {max_len_discards} samples longer than max length were "
                    f"found. The {max_len_discards} samples will be DISCARDED. If you are doing some kind of evaluation"
                    f", this can INVALIDATE results. This might happen if the max length was not set to -1 or if the "
                    f"sample length exceeds the maximum length supported by the current model."
                )
            else:
                logger.warning(
                    f"During iteration, {max_len_discards} elements were "
                    f"discarded since longer than max length {self.max_length}"
                )

        if min_len_discards > 0:
            if self.for_inference:
                logger.warning(
                    f"WARNING: Inference mode is True but {min_len_discards} samples shorter than min length were "
                    f"found. The {min_len_discards} samples will be DISCARDED. If you are doing some kind of evaluation"
                    f", this can INVALIDATE results. This might happen if the min length was not set to -1 or if the "
                    f"sample length is shorter than the minimum length supported by the current model."
                )
            else:
                logger.warning(
                    f"During iteration, {min_len_discards} elements were "
                    f"discarded since shorter than min length {self.min_length}"
                )

    @staticmethod
    def _new_output_format(sample: RelikReaderSample) -> RelikReaderSample:
        # try-out for a new format

        # set of span tuples (start, end, type) for each entity
        predicted_spans = set()
        for prediction in sample.predicted_entities:
            predicted_spans.add(
                (
                    prediction[0],
                    prediction[1],
                    prediction[2],
                )
            )

        # sort the spans by start so that we can use the index of the span to get the entity
        predicted_spans = sorted(predicted_spans, key=lambda x: x[0])
        predicted_triplets = []
        # now search for the spans in each triplet
        for prediction in sample.predicted_relations:
            # get the index of the entity that has the same start and end
            start_entity_index = [
                i
                for i, p in enumerate(predicted_spans)
                if p[:2] == (prediction["subject"][0], prediction["subject"][1])
            ][0]
            end_entity_index = [
                i
                for i, p in enumerate(predicted_spans)
                if p[:2] == (prediction["object"][0], prediction["object"][1])
            ][0]

            predicted_triplets.append(
                (
                    start_entity_index,
                    prediction["relation"],
                    end_entity_index,
                    prediction["probability"],
                )
            )
        sample.predicted_spans = predicted_spans
        sample.predicted_triplets = predicted_triplets
        return sample

    @staticmethod
    def _convert_annotations(sample: RelikReaderSample) -> RelikReaderSample:
        triplets = []
        entities = []

        for entity in sample.predicted_entities:
            span_start = entity[0] - 1
            span_end = entity[1] - 1
            if str(span_start) not in sample.token2word_start:
                # span_start is in the middle of a word
                # retrieve the first token of the word
                while str(span_start) not in sample.token2word_start:
                    span_start -= 1
                    # skip
                    if span_start < 0:
                        break
            if str(span_end) not in sample.token2word_end:
                # span_end is in the middle of a word
                # retrieve the last token of the word
                while str(span_end) not in sample.token2word_end:
                    span_end += 1
                    # skip
                    if span_end >= len(sample.tokens):
                        break

            if span_start < 0 or span_end >= len(sample.tokens):
                continue

            entities.append(
                (
                    sample.token2word_start[str(span_start)],
                    sample.token2word_end[str(span_end)] + 1,
                    (
                        sample.span_candidates[entity[2]]
                        if sample.span_candidates and len(entity) > 2
                        else "--NME--"
                    ),
                )
            )
        for predicted_triplet, predicted_triplet_probabilities in zip(
            sample.predicted_relations, sample.predicted_relations_probabilities
        ):
            subject, object_, relation = predicted_triplet
            subject = entities[subject]
            object_ = entities[object_]
            relation = sample.triplet_candidates[relation]
            triplets.append(
                {
                    "subject": subject,
                    "relation": relation,
                    "object": object_,
                    "probability": float(predicted_triplet_probabilities.round(2)),
                }
            )
        # convert to list since we need to modify the sample down the road
        sample.predicted_entities = entities
        sample.predicted_relations = triplets
        del sample._d["predicted_relations_probabilities"]

        return sample

    @staticmethod
    def convert_to_word_annotations(sample: RelikReaderSample) -> RelikReaderSample:
        sample = RelikREDataset._convert_annotations(sample)
        return RelikREDataset._new_output_format(sample)

    @staticmethod
    def convert_to_char_annotations(
        sample: RelikReaderSample,
        remove_nmes: bool = True,
    ) -> RelikReaderSample:
        RelikREDataset._convert_annotations(sample)
        if "token2char_start" in sample._d:
            entities = []
            for entity in sample.predicted_entities:
                entity = list(entity)
                token_start = sample.word2token_start[str(entity[0])]
                entity[0] = sample.token2char_start[str(token_start)]
                token_end = sample.word2token_end[str(entity[1] - 1)]
                entity[1] = sample.token2char_end[str(token_end)]
                entities.append(entity)
            sample.predicted_entities = entities
            for triplet in sample.predicted_relations:
                triplet["subject"] = (
                    sample.token2char_start[
                        str(sample.word2token_start[str(triplet["subject"][0])])
                    ],
                    sample.token2char_end[
                        str(sample.word2token_end[str(triplet["subject"][1] - 1)])
                    ],
                    triplet["subject"][2],
                )
                triplet["object"] = (
                    sample.token2char_start[
                        str(sample.word2token_start[str(triplet["object"][0])])
                    ],
                    sample.token2char_end[
                        str(sample.word2token_end[str(triplet["object"][1] - 1)])
                    ],
                    triplet["object"][2],
                )
                # triplet["object"][0] = sample.token2char_start[
                #     str(sample.word2token_start[str(triplet["object"][0])])
                # ]
                # triplet["object"][1] = sample.token2char_end[
                #     str(sample.word2token_end[str(triplet["object"][1] - 1)])
                # ]

            sample = RelikREDataset._new_output_format(sample)

        return sample

    @staticmethod
    def merge_patches_predictions(sample) -> None:
        pass


def main():
    special_symbols = [NME_SYMBOL] + [f"R-{i}" for i in range(50)]

    relik_dataset = RelikREDataset(
        "/home/huguetcabot/alby-re/alby/data/nyt-alby+/valid.jsonl",
        materialize_samples=False,
        transformer_model="microsoft/deberta-v3-base",
        special_symbols=special_symbols,
        shuffle_candidates=False,
        flip_candidates=False,
        for_inference=True,
    )

    for batch in relik_dataset:
        print(batch)
        exit(0)


if __name__ == "__main__":
    main()



# from relik.reader.pytorch_modules.base import RelikReaderBase
# from relik.reader.pytorch_modules.span import RelikReaderForSpanExtraction
# from relik.reader.pytorch_modules.triplet import RelikReaderForTripletExtraction



import json
import numpy as np
from typing import Iterable


class NpEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, np.integer):
            return int(obj)
        if isinstance(obj, np.floating):
            return float(obj)
        if isinstance(obj, np.ndarray):
            return obj.tolist()
        return super(NpEncoder, self).default(obj)


class RelikReaderSample:
    def __init__(self, **kwargs):
        super().__setattr__("_d", {})
        self._d = kwargs

    def __getattribute__(self, item):
        return super(RelikReaderSample, self).__getattribute__(item)

    def __getattr__(self, item):
        if item.startswith("__") and item.endswith("__"):
            # this is likely some python library-specific variable (such as __deepcopy__ for copy)
            # better follow standard behavior here
            raise AttributeError(item)
        elif item in self._d:
            return self._d[item]
        else:
            return None

    def __setattr__(self, key, value):
        if key in self._d:
            self._d[key] = value
        else:
            super().__setattr__(key, value)

    def to_jsons(self) -> str:
        if "predicted_window_labels" in self._d:
            new_obj = {
                k: v
                for k, v in self._d.items()
                if k != "predicted_window_labels" and k != "span_title_probabilities"
            }
            new_obj["predicted_window_labels"] = [
                [ss, se, pred_title]
                for (ss, se), pred_title in self.predicted_window_labels_chars
            ]
        else:
            return json.dumps(self._d, cls=NpEncoder)

    def to_dict(self) -> dict:
        return self._d


def load_relik_reader_samples(path: str) -> Iterable[RelikReaderSample]:
    with open(path) as f:
        for line in f:
            jsonl_line = json.loads(line.strip())
            relik_reader_sample = RelikReaderSample(**jsonl_line)
            yield relik_reader_sample



from typing import Dict, List

from lightning.pytorch.callbacks import Callback

from relik.reader.data.relik_reader_sample import RelikReaderSample
from relik.reader.utils.relik_reader_predictor import RelikReaderPredictor
from relik.reader.utils.metrics import f1_measure, safe_divide
from relik.reader.utils.special_symbols import NME_SYMBOL


class StrongMatching:
    def __call__(self, predicted_samples: List[RelikReaderSample]) -> Dict:
        # accumulators
        correct_predictions = 0
        correct_predictions_at_k = 0
        total_predictions = 0
        total_gold = 0
        correct_span_predictions = 0
        miss_due_to_candidates = 0

        # prediction index stats
        avg_correct_predicted_index = []
        avg_wrong_predicted_index = []
        less_index_predictions = []

        # collect data from samples
        for sample in predicted_samples:
            predicted_annotations = sample.predicted_window_labels_chars
            predicted_annotations_probabilities = sample.probs_window_labels_chars
            gold_annotations = {
                (ss, se, entity)
                for ss, se, entity in sample.window_labels
                if entity != NME_SYMBOL
            }
            total_predictions += len(predicted_annotations)
            total_gold += len(gold_annotations)

            # correct named entity detection
            predicted_spans = {(s, e) for s, e, _ in predicted_annotations}
            gold_spans = {(s, e) for s, e, _ in gold_annotations}
            correct_span_predictions += len(predicted_spans.intersection(gold_spans))

            # correct entity linking
            correct_predictions += len(
                predicted_annotations.intersection(gold_annotations)
            )

            for ss, se, ge in gold_annotations.difference(predicted_annotations):
                if ge not in sample.span_candidates:
                    miss_due_to_candidates += 1
                if ge in predicted_annotations_probabilities.get((ss, se), set()):
                    correct_predictions_at_k += 1

            # indices metrics
            predicted_spans_index = {
                (ss, se): ent for ss, se, ent in predicted_annotations
            }
            gold_spans_index = {(ss, se): ent for ss, se, ent in gold_annotations}

            for pred_span, pred_ent in predicted_spans_index.items():
                gold_ent = gold_spans_index.get(pred_span)

                if pred_span not in gold_spans_index:
                    continue

                # missing candidate
                if gold_ent not in sample.span_candidates:
                    continue

                gold_idx = sample.span_candidates.index(gold_ent)
                if gold_idx is None:
                    continue
                pred_idx = sample.span_candidates.index(pred_ent)

                if gold_ent != pred_ent:
                    avg_wrong_predicted_index.append(pred_idx)

                    if gold_idx is not None:
                        if pred_idx > gold_idx:
                            less_index_predictions.append(0)
                        else:
                            less_index_predictions.append(1)

                else:
                    avg_correct_predicted_index.append(pred_idx)

        # compute NED metrics
        span_precision = safe_divide(correct_span_predictions, total_predictions)
        span_recall = safe_divide(correct_span_predictions, total_gold)
        span_f1 = f1_measure(span_precision, span_recall)

        # compute EL metrics
        precision = safe_divide(correct_predictions, total_predictions)
        recall = safe_divide(correct_predictions, total_gold)
        recall_at_k = safe_divide(
            (correct_predictions + correct_predictions_at_k), total_gold
        )

        f1 = f1_measure(precision, recall)

        wrong_for_candidates = safe_divide(miss_due_to_candidates, total_gold)

        out_dict = {
            "span_precision": span_precision,
            "span_recall": span_recall,
            "span_f1": span_f1,
            "core_precision": precision,
            "core_recall": recall,
            "core_recall-at-k": recall_at_k,
            "core_f1": round(f1, 4),
            "wrong-for-candidates": wrong_for_candidates,
            "index_errors_avg-index": safe_divide(
                sum(avg_wrong_predicted_index), len(avg_wrong_predicted_index)
            ),
            "index_correct_avg-index": safe_divide(
                sum(avg_correct_predicted_index), len(avg_correct_predicted_index)
            ),
            "index_avg-index": safe_divide(
                sum(avg_correct_predicted_index + avg_wrong_predicted_index),
                len(avg_correct_predicted_index + avg_wrong_predicted_index),
            ),
            "index_percentage-favoured-smaller-idx": safe_divide(
                sum(less_index_predictions), len(less_index_predictions)
            ),
        }

        return {k: round(v, 5) for k, v in out_dict.items()}


class ELStrongMatchingCallback(Callback):
    def __init__(self, dataset_path: str, dataset_conf) -> None:
        super().__init__()
        self.dataset_path = dataset_path
        self.dataset_conf = dataset_conf
        self.strong_matching_metric = StrongMatching()

    def on_validation_epoch_start(self, trainer, pl_module) -> None:
        relik_reader_predictor = RelikReaderPredictor(pl_module.relik_reader_core_model)
        predicted_samples = relik_reader_predictor.predict(
            self.dataset_path,
            samples=None,
            dataset_conf=self.dataset_conf,
        )
        predicted_samples = list(predicted_samples)
        for k, v in self.strong_matching_metric(predicted_samples).items():
            pl_module.log(f"val_{k}", v)



import contextlib
import logging
import os
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Callable, Dict, List, Optional, Union

import numpy
import psutil
import torch
from omegaconf import OmegaConf
from pprintpp import pformat
from torch.utils.data import DataLoader
from tqdm import tqdm

from relik.common.log import get_logger
from relik.common.upload import upload
from relik.common.utils import from_cache, is_package_available
from relik.retriever.common.model_inputs import ModelInputs
from relik.retriever.data.base.datasets import BaseDataset
from relik.retriever.indexers.base import BaseDocumentIndex
from relik.retriever.indexers.document import Document, DocumentStore
from relik.retriever.pytorch_modules import PRECISION_MAP, RetrievedSample

if is_package_available("faiss"):
    import faiss
    import faiss.contrib.torch_utils

logger = get_logger(__name__, level=logging.INFO)


@dataclass
class FaissOutput:
    indices: Union[torch.Tensor, numpy.ndarray]
    distances: Union[torch.Tensor, numpy.ndarray]


class FaissDocumentIndex(BaseDocumentIndex):
    DOCUMENTS_FILE_NAME = "documents.jsonl"
    EMBEDDINGS_FILE_NAME = "embeddings.pt"
    INDEX_FILE_NAME = "index.faiss"

    def __init__(
        self,
        documents: (
            str | List[str] | os.PathLike | List[os.PathLike] | DocumentStore | None
        ) = None,
        embeddings: torch.Tensor | numpy.ndarray | None = None,
        metadata_fields: List[str] | None = None,
        separator: str | None = None,
        name_or_path: str | os.PathLike | None = None,
        device: str = "cpu",
        index=None,
        index_type: str = "Flat",
        metric: int = 0,  # -> faiss.METRIC_INNER_PRODUCT
        normalize: bool = False,
        *args,
        **kwargs,
    ) -> None:
        super().__init__(
            documents, embeddings, metadata_fields, separator, name_or_path, device
        )

        if embeddings is not None and documents is not None:
            logger.info("Both documents and embeddings are provided.")
            if len(documents) != embeddings.shape[0]:
                raise ValueError(
                    "The number of documents and embeddings must be the same. "
                    f"Got {len(documents)} documents and {embeddings.shape[0]} embeddings."
                )

        # faiss.omp_set_num_threads(psutil.cpu_count(logical=True))

        # params
        self.index_type = index_type
        self.metric = metric
        self.normalize = normalize

        if index is not None:
            self.embeddings = index
            if "cuda" in self.device:
                # use a single GPU
                # infer the device number if present
                device_number = (
                    int(self.device.split(":")[-1]) if ":" in self.device else 0
                )
                faiss_resource = faiss.StandardGpuResources()
                self.embeddings = faiss.index_cpu_to_gpu(
                    faiss_resource, device_number, self.embeddings
                )
        else:
            if embeddings is not None:
                # build the faiss index
                logger.info("Building the index from the embeddings.")
                self.embeddings = self._build_faiss_index(
                    embeddings=embeddings,
                    index_type=index_type,
                    normalize=normalize,
                    metric=metric,
                )

    def to(
        self, device_or_precision: str | torch.device | torch.dtype
    ) -> "BaseDocumentIndex":
        """
        Move the retriever to the specified device or precision.

        Args:
            device_or_precision (`str` | `torch.device` | `torch.dtype`):
                The device or precision to move the retriever to.

        Returns:
            `BaseDocumentIndex`: The retriever.
        """
        if isinstance(device_or_precision, torch.dtype):
            raise ValueError(
                "FaissDocumentIndex does not support precision conversion."
            )
        if device_or_precision == "cuda" and self.device == "cpu":
            # use a single GPU
            device_number = (
                int(device_or_precision.split(":")[-1])
                if ":" in device_or_precision
                else 0
            )
            faiss_resource = faiss.StandardGpuResources()
            self.embeddings = faiss.index_cpu_to_gpu(
                faiss_resource, device_number, self.embeddings
            )
        elif device_or_precision == "cpu" and self.device == "cuda":
            # move faiss index to CPU
            self.embeddings = faiss.index_gpu_to_cpu(self.embeddings)
        else:
            logger.warning(
                f"Provided device `{device_or_precision}` is the same as the current device `{self.device}`."
            )
        return self

    @property
    def device(self):
        # check if faiss index is on GPU
        if faiss.get_num_gpus() > 0:
            return "cuda"
        return "cpu"

    def _build_faiss_index(
        self,
        embeddings: Optional[Union[torch.Tensor, numpy.ndarray]],
        index_type: str,
        normalize: bool,
        metric: int,
    ):
        # build the faiss index
        self.normalize = (
            normalize
            and metric == 0  # -> faiss.METRIC_INNER_PRODUCT
            and not isinstance(embeddings, torch.Tensor)
        )
        if self.normalize:
            index_type = f"L2norm,{index_type}"
        faiss_vector_size = embeddings.shape[1]
        # if self.device == "cpu":
        #     index_type = index_type.replace("x,", "x_HNSW32,")
        # nlist = math.ceil(math.sqrt(faiss_vector_size)) * 4
        # # nlist = 8
        # index_type = index_type.replace(
        #     "x", str(nlist)
        # )
        # print("Current nlist:", nlist)
        self.embeddings = faiss.index_factory(faiss_vector_size, index_type, metric)

        # convert to GPU
        if "cuda" in self.device:
            # move the index to the GPU
            self.to(self.device)
        else:
            # move the tensors to the CPU
            embeddings = (
                embeddings.cpu() if isinstance(embeddings, torch.Tensor) else embeddings
            )

        # convert to float32 if embeddings is a torch.Tensor and not already float32
        if isinstance(embeddings, torch.Tensor) and embeddings.dtype != torch.float32:
            embeddings = embeddings.float()

        # logger.info("Training the index.")
        # self.embeddings.train(embeddings)

        logger.info("Adding the embeddings to the index.")
        self.embeddings.add(embeddings)

        # self.embeddings.nprobe = nprobe

        # save parameters for saving/loading
        self.index_type = index_type
        self.metric = metric

        # clear the embeddings to free up memory
        embeddings = None

        return self.embeddings

    @torch.no_grad()
    @torch.inference_mode()
    def index(
        self,
        retriever,
        documents: Optional[List[Document]] = None,
        batch_size: int = 32,
        num_workers: int = 4,
        max_length: int | None = None,
        collate_fn: Optional[Callable] = None,
        encoder_precision: Optional[Union[str, int]] = None,
        compute_on_cpu: bool = False,
        force_reindex: bool = False,
        *args,
        **kwargs,
    ) -> "FaissDocumentIndex":
        """
        Index the documents using the encoder.

        Args:
            retriever (:obj:`torch.nn.Module`):
                The encoder to be used for indexing.
            documents (:obj:`List[Document]`, `optional`, defaults to None):
                The documents to be indexed.
            batch_size (:obj:`int`, `optional`, defaults to 32):
                The batch size to be used for indexing.
            num_workers (:obj:`int`, `optional`, defaults to 4):
                The number of workers to be used for indexing.
            max_length (:obj:`int`, `optional`, defaults to None):
                The maximum length of the input to the encoder.
            collate_fn (:obj:`Callable`, `optional`, defaults to None):
                The collate function to be used for batching.
            encoder_precision (:obj:`Union[str, int]`, `optional`, defaults to None):
                The precision to be used for the encoder.
            compute_on_cpu (:obj:`bool`, `optional`, defaults to False):
                Whether to compute the embeddings on CPU.
            force_reindex (:obj:`bool`, `optional`, defaults to False):
                Whether to force reindexing.

        Returns:
            :obj:`InMemoryIndexer`: The indexer object.
        """

        if self.embeddings is not None and not force_reindex:
            logger.info(
                "Embeddings are already present and `force_reindex` is `False`. Skipping indexing."
            )
            if documents is None:
                return self

        # release the memory
        if collate_fn is None:
            tokenizer = retriever.passage_tokenizer

            def collate_fn(x):
                return ModelInputs(
                    tokenizer(
                        x,
                        padding=True,
                        return_tensors="pt",
                        truncation=True,
                        max_length=max_length or tokenizer.model_max_length,
                    )
                )

        if force_reindex:
            if documents is not None:
                self.documents.add_document(documents)
            data = [k for k in self.get_passages()]

        else:
            if documents is not None:
                data = [k for k in self.get_passages(DocumentStore(documents))]
            else:
                return self

        dataloader = DataLoader(
            BaseDataset(name="passage", data=data),
            batch_size=batch_size,
            shuffle=False,
            num_workers=num_workers,
            pin_memory=False,
            collate_fn=collate_fn,
        )

        encoder = retriever.passage_encoder

        # Create empty lists to store the passage embeddings and passage index
        passage_embeddings: List[torch.Tensor] = []

        encoder_device = "cpu" if compute_on_cpu else self.device

        # fucking autocast only wants pure strings like 'cpu' or 'cuda'
        # we need to convert the model device to that
        device_type_for_autocast = str(encoder_device).split(":")[0]
        # autocast doesn't work with CPU and stuff different from bfloat16
        autocast_pssg_mngr = (
            contextlib.nullcontext()
            if device_type_for_autocast == "cpu"
            else (
                torch.autocast(
                    device_type=device_type_for_autocast,
                    dtype=PRECISION_MAP[encoder_precision],
                )
            )
        )
        with autocast_pssg_mngr:
            # Iterate through each batch in the dataloader
            for batch in tqdm(dataloader, desc="Indexing"):
                # Move the batch to the device
                batch: ModelInputs = batch.to(encoder_device)
                # Compute the passage embeddings
                passage_outs = encoder(**batch)
                # Append the passage embeddings to the list
                if self.device == "cpu":
                    passage_embeddings.extend([c.detach().cpu() for c in passage_outs])
                else:
                    passage_embeddings.extend([c for c in passage_outs])

        # move the passage embeddings to the CPU if not already done
        passage_embeddings = [c.detach().cpu() for c in passage_embeddings]
        # stack it
        passage_embeddings: torch.Tensor = torch.stack(passage_embeddings, dim=0)
        # convert to float32 for faiss
        passage_embeddings.to(PRECISION_MAP["float32"])

        # index the embeddings
        self.embeddings = self._build_faiss_index(
            embeddings=passage_embeddings,
            index_type=self.index_type,
            normalize=self.normalize,
            metric=self.metric,
        )
        # free up memory from the unused variable
        del passage_embeddings

        return self

    @torch.no_grad()
    @torch.inference_mode()
    def search(self, query: torch.Tensor, k: int = 1) -> list[list[RetrievedSample]]:
        k = min(k, self.embeddings.ntotal)

        if self.normalize:
            faiss.normalize_L2(query)
        if isinstance(query, torch.Tensor) and self.device == "cpu":
            query = query.detach().cpu()
        # Retrieve the indices of the top k passage embeddings
        retriever_out = self.embeddings.search(query, k)

        # get int values (second element of the tuple)
        batch_top_k: List[List[int]] = retriever_out[1].detach().cpu().tolist()
        # get float values (first element of the tuple)
        batch_scores: List[List[float]] = retriever_out[0].detach().cpu().tolist()
        # Retrieve the passages corresponding to the indices
        batch_docs = [
            [self.get_document_from_index(i) for i in indices if i != -1]
            for indices in batch_top_k
        ]
        # build the output object
        batch_retrieved_samples = [
            [
                RetrievedSample(document=doc, score=score)
                for doc, score in zip(docs, scores)
            ]
            for docs, scores in zip(batch_docs, batch_scores)
        ]
        return batch_retrieved_samples

    def save_pretrained(
        self,
        output_dir: Union[str, os.PathLike],
        config: Optional[Dict[str, Any]] = None,
        config_file_name: str | None = None,
        document_file_name: str | None = None,
        embedding_file_name: str | None = None,
        index_file_name: str | None = None,
        push_to_hub: bool = False,
        model_id: str | None = None,
        **kwargs,
    ):
        """
        Save the retriever to a directory.

        Args:
            output_dir (`str`):
                The directory to save the retriever to.
            config (`Optional[Dict[str, Any]]`, `optional`):
                The configuration to save. If `None`, the current configuration of the retriever will be
                saved. Defaults to `None`.
            config_file_name (`str | None`, `optional`):
                The name of the configuration file. Defaults to `config.yaml`.
            document_file_name (`str | None`, `optional`):
                The name of the document file. Defaults to `documents.json`.
            embedding_file_name (`str | None`, `optional`):
                The name of the embedding file. Defaults to `embeddings.pt`.
            push_to_hub (`bool`, `optional`):
                Whether to push the saved retriever to the hub. Defaults to `False`.
            model_id (`str | None`, `optional`):
                The id of the model to push to the hub. If `None`, the name of the output
                directory will be used. Defaults to `None`.
            **kwargs:
                Additional keyword arguments to pass to `upload`.
        """
        if config is None:
            # create a default config
            config = self.config

        config_file_name = config_file_name or self.CONFIG_NAME
        document_file_name = document_file_name or self.DOCUMENTS_FILE_NAME
        index_file_name = index_file_name or self.INDEX_FILE_NAME

        # create the output directory
        output_dir = Path(output_dir)
        output_dir.mkdir(parents=True, exist_ok=True)

        logger.info(f"Saving retriever to {output_dir}")
        logger.info(f"Saving config to {output_dir / config_file_name}")
        # pretty print the config
        OmegaConf.save(config, output_dir / config_file_name)
        logger.info(pformat(config))

        # save the current state of the retriever
        index_path = output_dir / index_file_name
        logger.info(f"Saving FAISS index state to {output_dir / index_path}")
        faiss.write_index(self.embeddings, str(index_path))

        # save the passage index
        documents_path = output_dir / document_file_name
        logger.info(f"Saving passage index to {documents_path}")
        self.documents.save(documents_path)

        logger.info("Saving document index to disk done.")

        if push_to_hub:
            # push to hub
            logger.info("Pushing to hub")
            model_id = model_id or output_dir.name
            upload(output_dir, model_id, **kwargs)

    def get_embeddings_from_index(
        self, index: int
    ) -> Union[torch.Tensor, numpy.ndarray]:
        """
        Get the document vector from the index.

        Args:
            index (`int`):
                The index of the document.

        Returns:
            `torch.Tensor`: The document vector.
        """
        if self.embeddings is None:
            raise ValueError(
                "The documents must be indexed before they can be retrieved."
            )
        if index >= self.embeddings.ntotal:
            raise ValueError(
                f"The index {index} is out of bounds. The maximum index is {self.embeddings.ntotal}."
            )
        return self.embeddings.reconstruct(index)



from typing import List, Union


class BlankSentenceSplitter:
    """
    A `BlankSentenceSplitter` splits strings into sentences.
    """

    def __call__(self, *args, **kwargs):
        """
        Calls :meth:`split_sentences`.
        """
        return self.split_sentences(*args, **kwargs)

    def split_sentences(
        self, text: str, max_len: int = 0, *args, **kwargs
    ) -> List[str]:
        """
        Splits a `text` :class:`str` paragraph into a list of :class:`str`, where each is a sentence.
        """
        return [text]

    def split_sentences_batch(
        self, texts: List[str], *args, **kwargs
    ) -> List[List[str]]:
        """
        Default implementation is to just iterate over the texts and call `split_sentences`.
        """
        return [self.split_sentences(text) for text in texts]



import contextlib
import logging
from typing import Any, Dict, Iterator, List

import numpy as np
import torch
import transformers as tr
from lightning_fabric.utilities import move_data_to_device
from torch.utils.data import DataLoader, IterableDataset
from tqdm import tqdm

from relik.common.log import get_logger
from relik.common.torch_utils import get_autocast_context
from relik.common.utils import get_callable_from_string
from relik.inference.data.objects import AnnotationType
from relik.reader.data.relik_reader_sample import RelikReaderSample
from relik.reader.pytorch_modules.base import RelikReaderBase
from relik.retriever.pytorch_modules import PRECISION_MAP

logger = get_logger(__name__, level=logging.INFO)


class RelikReaderForTripletExtraction(RelikReaderBase):
    """
    A class for the RelikReader model for triplet extraction.

    Args:
        transformer_model (:obj:`str` or :obj:`transformers.PreTrainedModel` or :obj:`None`, `optional`):
            The transformer model to use. If `None`, the default model is used.
        additional_special_symbols (:obj:`int`, `optional`, defaults to 0):
            The number of additional special symbols to add to the tokenizer.
        num_layers (:obj:`int`, `optional`):
            The number of layers to use. If `None`, all layers are used.
        activation (:obj:`str`, `optional`, defaults to "gelu"):
            The activation function to use.
        linears_hidden_size (:obj:`int`, `optional`, defaults to 512):
            The hidden size of the linears.
        use_last_k_layers (:obj:`int`, `optional`, defaults to 1):
            The number of last layers to use.
        training (:obj:`bool`, `optional`, defaults to False):
            Whether the model is in training mode.
        device (:obj:`str` or :obj:`torch.device` or :obj:`None`, `optional`):
            The device to use. If `None`, the default device is used.
        tokenizer (:obj:`str` or :obj:`transformers.PreTrainedTokenizer` or :obj:`None`, `optional`):
            The tokenizer to use. If `None`, the default tokenizer is used.
        dataset (:obj:`IterableDataset` or :obj:`str` or :obj:`None`, `optional`):
            The dataset to use. If `None`, the default dataset is used.
        dataset_kwargs (:obj:`Dict[str, Any]` or :obj:`None`, `optional`):
            The keyword arguments to pass to the dataset class.
        default_reader_class (:obj:`str` or :obj:`transformers.PreTrainedModel` or :obj:`None`, `optional`):
            The default reader class to use. If `None`, the default reader class is used.
        **kwargs:
            Keyword arguments.
    """

    default_reader_class: str = (
        "relik.reader.pytorch_modules.hf.modeling_relik.RelikReaderREModel"
    )
    default_data_class: str = "relik.reader.data.relik_reader_re_data.RelikREDataset"

    def __init__(
        self,
        transformer_model: str | tr.PreTrainedModel | None = None,
        additional_special_symbols: int = 0,
        additional_special_symbols_types: int = 0,
        entity_type_loss: bool | None = None,
        add_entity_embedding: bool | None = None,
        num_layers: int | None = None,
        activation: str = "gelu",
        linears_hidden_size: int | None = 512,
        use_last_k_layers: int = 1,
        training: bool = False,
        device: str | torch.device | None = None,
        tokenizer: str | tr.PreTrainedTokenizer | None = None,
        dataset: IterableDataset | str | None = None,
        dataset_kwargs: Dict[str, Any] | None = None,
        default_reader_class: tr.PreTrainedModel | str | None = None,
        **kwargs,
    ):
        super().__init__(
            transformer_model=transformer_model,
            additional_special_symbols=additional_special_symbols,
            additional_special_symbols_types=additional_special_symbols_types,
            entity_type_loss=entity_type_loss,
            add_entity_embedding=add_entity_embedding,
            num_layers=num_layers,
            activation=activation,
            linears_hidden_size=linears_hidden_size,
            use_last_k_layers=use_last_k_layers,
            training=training,
            device=device,
            tokenizer=tokenizer,
            dataset=dataset,
            default_reader_class=default_reader_class,
            **kwargs,
        )
        # and instantiate the dataset class
        self.dataset = dataset
        if self.dataset is None and training is False:
            self.default_data_class = get_callable_from_string(self.default_data_class)
            default_data_kwargs = dict(
                dataset_path=None,
                materialize_samples=False,
                transformer_model=self.tokenizer,
                special_symbols_re=self.default_data_class.get_special_symbols_re(
                    self.relik_reader_model.config.additional_special_symbols,
                    use_nme=kwargs.get("use_nme_re", False),
                ),
                special_symbols=(
                    self.default_data_class.get_special_symbols(
                        self.relik_reader_model.config.additional_special_symbols_types
                        - 1
                    )
                    if self.relik_reader_model.config.additional_special_symbols_types
                    > 0
                    else []
                ),
                for_inference=True,
                use_nme=kwargs.get("use_nme", False),
            )
            # merge the default data kwargs with the ones passed to the model
            default_data_kwargs.update(dataset_kwargs or {})
            self.dataset = self.default_data_class(**default_data_kwargs)

    @torch.no_grad()
    @torch.inference_mode()
    def _read(
        self,
        samples: List[RelikReaderSample] | None = None,
        input_ids: torch.Tensor | None = None,
        attention_mask: torch.Tensor | None = None,
        token_type_ids: torch.Tensor | None = None,
        prediction_mask: torch.Tensor | None = None,
        special_symbols_mask: torch.Tensor | None = None,
        max_length: int = 2048,
        max_batch_size: int = 128,
        token_batch_size: int = 2048,
        precision: str = 32,
        annotation_type: AnnotationType = AnnotationType.WORD,
        progress_bar: bool = False,
        *args: object,
        **kwargs: object,
    ) -> List[RelikReaderSample] | List[List[RelikReaderSample]]:
        """
        A wrapper around the forward method that returns the predicted labels for each sample.

        Args:
            samples (:obj:`List[RelikReaderSample]`, `optional`):
                The samples to read. If provided, `text` and `candidates` are ignored.
            input_ids (:obj:`torch.Tensor`, `optional`):
                The input ids of the text. If `samples` is provided, this is ignored.
            attention_mask (:obj:`torch.Tensor`, `optional`):
                The attention mask of the text. If `samples` is provided, this is ignored.
            token_type_ids (:obj:`torch.Tensor`, `optional`):
                The token type ids of the text. If `samples` is provided, this is ignored.
            prediction_mask (:obj:`torch.Tensor`, `optional`):
                The prediction mask of the text. If `samples` is provided, this is ignored.
            special_symbols_mask (:obj:`torch.Tensor`, `optional`):
                The special symbols mask of the text. If `samples` is provided, this is ignored.
            max_length (:obj:`int`, `optional`, defaults to 1000):
                The maximum length of the text.
            max_batch_size (:obj:`int`, `optional`, defaults to 128):
                The maximum batch size.
            token_batch_size (:obj:`int`, `optional`):
                The token batch size.
            progress_bar (:obj:`bool`, `optional`, defaults to False):
                Whether to show a progress bar.
            precision (:obj:`str`, `optional`, defaults to 32):
                The precision to use for the model.
            annotation_type (`AnnotationType`, `optional`, defaults to `AnnotationType.CHAR`):
                The type of annotation to return. If `char`, the spans will be in terms of
                character offsets. If `word`, the spans will be in terms of word offsets.
            *args:
                Positional arguments.
            **kwargs:
                Keyword arguments.

        Returns:
            :obj:`List[RelikReaderSample]` or :obj:`List[List[RelikReaderSample]]`:
                The predicted labels for each sample.
        """

        precision = precision or self.precision
        if samples is not None:

            def _read_iterator():
                def samples_it():
                    for i, sample in enumerate(samples):
                        assert sample._mixin_prediction_position is None
                        sample._mixin_prediction_position = i
                        if sample.spans is not None and len(sample.spans) > 0:
                            entities = []
                            offset_span = sample.char2token_start[str(sample.offset)]
                            for span_start, span_end in sample.spans:
                                entities.append(
                                    [
                                        span_start - offset_span,
                                        span_end + 1 - offset_span,
                                        "",
                                    ]
                                )
                                # if str(span_start) not in sample.char2token_start:
                                #     # span_start is in the middle of a word
                                #     # retrieve the first token of the word
                                #     while str(span_start) not in sample.char2token_start:
                                #         span_start -= 1
                                #         # skip
                                #         if span_start < 0:
                                #             break
                                # if str(span_end) not in sample.char2token_end:
                                #     # span_end is in the middle of a word
                                #     # retrieve the last token of the word
                                #     while str(span_end) not in sample.char2token_end:
                                #         span_end += 1
                                #         # skip
                                #         if span_end >= int(list(sample.char2token_end.keys())[-1]):
                                #             break

                                # if span_start < 0 or span_end > int(list(sample.char2token_end.keys())[-1]):
                                #     continue
                                # entities.append([sample.char2token_start[str(span_start)]-offset_span, sample.char2token_end[str(span_end)]+1-offset_span, ""])
                            sample.window_labels_tokens = entities
                        yield sample

                next_prediction_position = 0
                position2predicted_sample = {}

                # instantiate dataset
                if self.dataset is None:
                    raise ValueError(
                        "You need to pass a dataset to the model in order to predict"
                    )
                self.dataset.samples = samples_it()
                self.dataset.model_max_length = max_length
                self.dataset.tokens_per_batch = token_batch_size
                self.dataset.max_batch_size = max_batch_size

                # instantiate dataloader
                iterator = DataLoader(
                    self.dataset, batch_size=None, num_workers=0, shuffle=False
                )
                if progress_bar:
                    iterator = tqdm(iterator, desc="Predicting with RelikReader")

                with get_autocast_context(self.device, precision):
                    for batch in iterator:
                        batch = move_data_to_device(batch, self.device)
                        batch.update(kwargs)
                        batch_out = self._batch_predict(**batch)

                        for sample in batch_out:
                            if (
                                sample._mixin_prediction_position
                                >= next_prediction_position
                            ):
                                position2predicted_sample[
                                    sample._mixin_prediction_position
                                ] = sample

                        # yield
                        while next_prediction_position in position2predicted_sample:
                            yield position2predicted_sample[next_prediction_position]
                            del position2predicted_sample[next_prediction_position]
                            next_prediction_position += 1

            outputs = list(_read_iterator())
            for sample in outputs:
                self.dataset.merge_patches_predictions(sample)
                if annotation_type == AnnotationType.CHAR:
                    self.dataset.convert_to_char_annotations(sample)
                elif annotation_type == AnnotationType.WORD:
                    self.dataset.convert_to_word_annotations(sample)
                else:
                    raise ValueError(
                        f"Annotation type {annotation_type} not recognized. "
                        f"Please choose one of {list(AnnotationType)}."
                    )

        else:
            outputs = list(
                self._batch_predict(
                    input_ids,
                    attention_mask,
                    token_type_ids,
                    prediction_mask,
                    special_symbols_mask,
                    *args,
                    **kwargs,
                )
            )
        return outputs

    def _batch_predict(
        self,
        input_ids: torch.Tensor,
        attention_mask: torch.Tensor,
        token_type_ids: torch.Tensor | None = None,
        prediction_mask: torch.Tensor | None = None,
        special_symbols_mask: torch.Tensor | None = None,
        special_symbols_mask_entities: torch.Tensor | None = None,
        sample: List[RelikReaderSample] | None = None,
        *args,
        **kwargs,
    ) -> Iterator[RelikReaderSample]:
        """
        A wrapper around the forward method that returns the predicted labels for each sample.
        It also adds the predicted labels to the samples.

        Args:
            input_ids (:obj:`torch.Tensor`):
                The input ids of the text.
            attention_mask (:obj:`torch.Tensor`):
                The attention mask of the text.
            token_type_ids (:obj:`torch.Tensor`, `optional`):
                The token type ids of the text.
            prediction_mask (:obj:`torch.Tensor`, `optional`):
                The prediction mask of the text.
            special_symbols_mask (:obj:`torch.Tensor`, `optional`):
                The special symbols mask of the text.
            sample (:obj:`List[RelikReaderSample]`, `optional`):
                The samples to read. If provided, `text` and `candidates` are ignored.
            top_k (:obj:`int`, `optional`, defaults to 5):
                The amount of top-k most probable entities to predict.
            *args:
                Positional arguments.
            **kwargs:
                Keyword arguments.

        Returns:
            The predicted labels for each sample.
        """
        forward_output = self.forward(
            input_ids=input_ids,
            attention_mask=attention_mask,
            token_type_ids=token_type_ids,
            prediction_mask=prediction_mask,
            special_symbols_mask=special_symbols_mask,
            special_symbols_mask_entities=special_symbols_mask_entities,
            is_prediction=True,
            *args,
            **kwargs,
        )

        ned_start_predictions = forward_output["ned_start_predictions"].cpu().numpy()
        ned_end_predictions = forward_output["ned_end_predictions"]  # .cpu().numpy()
        ed_predictions = forward_output["re_entities_predictions"].cpu().numpy()
        ned_type_predictions = forward_output["ned_type_predictions"].cpu().numpy()
        re_predictions = forward_output["re_predictions"].cpu().numpy()
        re_probabilities = forward_output["re_probabilities"].detach().cpu().numpy()

        for ts, ne_st, ne_end, re_pred, re_prob, edp, ne_et in zip(
            sample,
            ned_start_predictions,
            ned_end_predictions,
            re_predictions,
            re_probabilities,
            ed_predictions,
            ned_type_predictions,
        ):
            ne_end = ne_end.cpu().numpy()
            entities = []
            if self.relik_reader_model.config.entity_type_loss:
                starts = np.argwhere(ne_st)
                i = 0
                for start, end in zip(starts, ne_end):
                    ends = np.argwhere(end)
                    for e in ends:
                        entities.append([start[0], e[0], ne_et[i]])
                        i += 1
                    #     if i == len(ne_et):
                    #         break
                    # if i == len(ne_et):
                    #     break
            else:
                starts = np.argwhere(ne_st)
                for start, end in zip(starts, ne_end):
                    ends = np.argwhere(end)
                    for e in ends:
                        entities.append([start[0], e[0]])

            edp = edp[: len(entities)]
            re_pred = re_pred[: len(entities), : len(entities)]
            re_prob = re_prob[: len(entities), : len(entities)]
            possible_re = np.argwhere(re_pred)
            predicted_triplets = []
            predicted_triplets_prob = []
            for i, j, r in possible_re:
                if self.relik_reader_model.relation_disambiguation_loss:
                    if not (
                        i != j
                        and edp[i, r] == 1
                        and edp[j, r] == 1
                        and edp[i, 0] == 0
                        and edp[j, 0] == 0
                    ):
                        continue
                predicted_triplets.append([i, j, r])
                predicted_triplets_prob.append(re_prob[i, j, r])

            ts._d["predicted_relations"] = predicted_triplets
            ts._d["predicted_entities"] = entities
            ts._d["predicted_relations_probabilities"] = predicted_triplets_prob
            if "return_threshold_utils" in kwargs and kwargs["return_threshold_utils"]:
                assert (
                    ts.window_triplet_labels_tokens is not None
                ), "The triplet labels are not provided, cannot return thresshold utils"
                ts._d["re_probabilities"] = re_prob
                # reconstruct the labels based on predicted_entities and ts.entities (gold)
                labels = np.zeros(re_prob.shape)
                for t in ts.window_triplet_labels_tokens:
                    i, j, r = t["subject"], t["object"], t["relation"]
                    subject_entity = [
                        ts.word2token_start[str(i[0])] + 1,
                        ts.word2token_end[str(i[1] - 1)] + 1,
                    ]
                    object_entity = [
                        ts.word2token_start[str(j[0])] + 1,
                        ts.word2token_end[str(j[1] - 1)] + 1,
                    ]
                    try:
                        subject_entity = entities.index(subject_entity)
                        object_entity = entities.index(object_entity)
                        relation_label = ts.triplet_candidates.index(r)
                        labels[subject_entity, object_entity, relation_label] = 1
                    except ValueError:
                        pass
                ts._d["re_labels"] = labels
            # try-out for a new format
            ts._d["predicted_triplets"] = predicted_triplets

            yield ts



from typing import List

import numpy as np
import torch


def flatten(lsts: List[list]) -> list:
    acc_lst = list()
    for lst in lsts:
        acc_lst.extend(lst)
    return acc_lst


def batchify(tensors: List[torch.Tensor], padding_value: int = 0) -> torch.Tensor:
    return torch.nn.utils.rnn.pad_sequence(
        tensors, batch_first=True, padding_value=padding_value
    )


def batchify_matrices(tensors: List[torch.Tensor], padding_value: int) -> torch.Tensor:
    x = max([t.shape[0] for t in tensors])
    y = max([t.shape[1] for t in tensors])
    out_matrix = torch.zeros((len(tensors), x, y))
    out_matrix += padding_value
    for i, tensor in enumerate(tensors):
        out_matrix[i][0 : tensor.shape[0], 0 : tensor.shape[1]] = tensor
    return out_matrix


def batchify_tensor(tensors: List[torch.Tensor], padding_value: int) -> torch.Tensor:
    x = max([t.shape[0] for t in tensors])
    y = max([t.shape[1] for t in tensors])
    rest = tensors[0].shape[2]
    out_matrix = torch.zeros((len(tensors), x, y, rest))
    out_matrix += padding_value
    for i, tensor in enumerate(tensors):
        out_matrix[i][0 : tensor.shape[0], 0 : tensor.shape[1], :] = tensor
    return out_matrix


def chunks(lst: list, chunk_size: int) -> List[list]:
    chunks_acc = list()
    for i in range(0, len(lst), chunk_size):
        chunks_acc.append(lst[i : i + chunk_size])
    return chunks_acc


def add_noise_to_value(value: int, noise_param: float):
    noise_value = value * noise_param
    noise = np.random.uniform(-noise_value, noise_value)
    return max(1, value + noise)



from argparse import ArgumentParser
import logging
import os
from pathlib import Path
from typing import Annotated, Dict, List, Union
import psutil

import torch
import uvicorn

from relik.common.utils import is_package_available
from relik.inference.annotator import Relik
from relik.inference.data.objects import AnnotationType, RelikOutput, TaskType
from relik.retriever.indexers.document import Document

if not is_package_available("fastapi"):
    raise ImportError(
        "FastAPI is not installed. Please install FastAPI with `pip install relik[serve]`."
    )
from fastapi import FastAPI, HTTPException, APIRouter, Query


from relik.common.log import get_logger
from relik.inference.serve.backend.utils import (
    RayParameterManager,
    ServerParameterManager,
)

logger = get_logger(__name__, level=logging.INFO)

VERSION = {}  # type: ignore
with open(
    Path(__file__).parent.parent.parent.parent / "version.py", "r"
) as version_file:
    exec(version_file.read(), VERSION)

# Env variables for server
SERVER_MANAGER = ServerParameterManager()
RAY_MANAGER = RayParameterManager()


class RelikServer:
    def __init__(
        self,
        relik_pretrained: str | None = None,
        device: str = "cpu",
        retriever_device: str | None = None,
        document_index_device: str | None = None,
        reader_device: str | None = None,
        precision: str | int | torch.dtype = 32,
        retriever_precision: str | int | torch.dtype | None = None,
        document_index_precision: str | int | torch.dtype | None = None,
        reader_precision: str | int | torch.dtype | None = None,
        annotation_type: str = "char",
        skip_metadata: bool = False,
        **kwargs,
    ):
        num_threads = os.getenv("TORCH_NUM_THREADS", psutil.cpu_count(logical=False))
        torch.set_num_threads(num_threads)
        logger.info(f"Torch is running on {num_threads} threads.")
        # parameters
        logger.info(f"RELIK_PRETRAINED: {relik_pretrained}")
        self.relik_pretrained = relik_pretrained
        logger.info(f"DEVICE: {device}")
        self.device = device
        if retriever_device is not None:
            logger.info(f"RETRIEVER_DEVICE: {retriever_device}")
        self.retriever_device = retriever_device or device
        if document_index_device is not None:
            logger.info(f"INDEX_DEVICE: {document_index_device}")
        self.document_index_device = document_index_device or retriever_device
        if reader_device is not None:
            logger.info(f"READER_DEVICE: {reader_device}")
        self.reader_device = reader_device
        logger.info(f"PRECISION: {precision}")
        self.precision = precision
        if retriever_precision is not None:
            logger.info(f"RETRIEVER_PRECISION: {retriever_precision}")
        self.retriever_precision = retriever_precision or precision
        if document_index_precision is not None:
            logger.info(f"INDEX_PRECISION: {document_index_precision}")
        self.document_index_precision = document_index_precision or precision
        if reader_precision is not None:
            logger.info(f"READER_PRECISION: {reader_precision}")
        self.reader_precision = reader_precision or precision
        logger.info(f"ANNOTATION_TYPE: {annotation_type}")
        self.annotation_type = annotation_type

        self.relik = Relik.from_pretrained(
            self.relik_pretrained,
            device=self.device,
            retriever_device=self.retriever_device,
            document_index_device=self.document_index_device,
            reader_device=self.reader_device,
            precision=self.precision,
            retriever_precision=self.retriever_precision,
            document_index_precision=self.document_index_precision,
            reader_precision=self.reader_precision,
            skip_metadata=skip_metadata,
        )

        self.router = APIRouter()
        self.router.add_api_route("/api/relik", self.relik_endpoint, methods=["GET"])

        logger.info("RelikServer initialized.")

    # @serve.batch()
    async def __call__(
        self,
        text: List[str],
        top_k: int | None = None,
        window_size: int | str | None = None,
        window_stride: int | str | None = None,
        is_split_into_words: bool = False,
        retriever_batch_size: int | None = 32,
        reader_batch_size: int | None = 32,
        return_windows: bool = False,
        use_doc_topic: bool = False,
        annotation_type: str | AnnotationType = AnnotationType.CHAR,
        relation_threshold: float = 0.5,
    ) -> List:
        output = self.relik(
            text=text,
            top_k=top_k,
            window_size=window_size,
            window_stride=window_stride,
            is_split_into_words=is_split_into_words,
            retriever_batch_size=retriever_batch_size,
            reader_batch_size=reader_batch_size,
            return_windows=return_windows,
            use_doc_topic=use_doc_topic,
            annotation_type=annotation_type,
            relation_threshold=relation_threshold,
        )
        output = output if isinstance(output, list) else [output]
        return [o.to_dict() for o in output]

    async def relik_endpoint(
        self,
        text: Annotated[list[str] | None, Query()] = None,
        top_k: int | None = None,
        window_size: int | str | None = None,
        window_stride: int | str | None = None,
        is_split_into_words: bool = False,
        retriever_batch_size: int | None = 32,
        reader_batch_size: int | None = 32,
        return_windows: bool = False,
        use_doc_topic: bool = False,
        annotation_type: str | AnnotationType = AnnotationType.CHAR,
        relation_threshold: float = 0.5,
    ) -> List:
        try:
            if window_size:
                # check if window size is a number as string
                if window_size.isdigit():
                    window_size = int(window_size)

            if window_stride:
                # check if window stride is a number as string
                if window_stride.isdigit():
                    window_stride = int(window_stride)

            # get predictions for the retriever
            return await self(
                text=text,
                top_k=top_k,
                window_size=window_size,
                window_stride=window_stride,
                is_split_into_words=is_split_into_words,
                retriever_batch_size=retriever_batch_size,
                reader_batch_size=reader_batch_size,
                return_windows=return_windows,
                use_doc_topic=use_doc_topic,
                annotation_type=annotation_type,
                relation_threshold=relation_threshold,
            )
        except Exception as e:
            # log the entire stack trace
            logger.exception(e)
            raise HTTPException(status_code=500, detail=f"Server Error: {e}")


def main(
    relik_pretrained: str,
    device: str = "cpu",
    retriever_device: str = None,
    document_index_device: str = None,
    reader_device: str = None,
    precision: str = "32",
    retriever_precision: str = None,
    document_index_precision: str = None,
    reader_precision: str = None,
    annotation_type: str = "char",
    workers: int = None,
    host: str = "localhost",
    port: int = 8000,
    frontend: bool = False,
):
    app = FastAPI(
        title="ReLiK - A blazing fast and lightweight Information Extraction model for Entity Linking and Relation Extraction.",
        version=VERSION["VERSION"],
        description="ReLiK REST API",
    )
    server = RelikServer(
        relik_pretrained=relik_pretrained,
        device=device,
        retriever_device=retriever_device,
        document_index_device=document_index_device,
        reader_device=reader_device,
        precision=precision,
        retriever_precision=retriever_precision,
        document_index_precision=document_index_precision,
        reader_precision=reader_precision,
        annotation_type=annotation_type,
    )
    app.include_router(server.router)
    if frontend:
        from relik.inference.serve.frontend.gradio_fe import main as serve_frontend
        import threading

        threading.Thread(target=serve_frontend, daemon=True).start()

    uvicorn.run(app, host=host, port=port, log_level="info", workers=workers)


if __name__ == "__main__":

    arg_parser = ArgumentParser()
    arg_parser.add_argument(
        "--relik_pretrained",
        type=str,
        help="Path to the pretrained ReLiK model",
    )
    arg_parser.add_argument(
        "--device", type=str, default="cpu", help="Device to run the model on"
    )
    arg_parser.add_argument(
        "--retriever_device",
        type=str,
        default=None,
        help="Device to run the retriever on",
    )
    arg_parser.add_argument(
        "--document_index_device",
        type=str,
        default=None,
        help="Device to run the document index on",
    )
    arg_parser.add_argument(
        "--reader_device", type=str, default=None, help="Device to run the reader on"
    )
    arg_parser.add_argument(
        "--precision", type=str, default=32, help="Precision of the model"
    )
    arg_parser.add_argument(
        "--retriever_precision",
        type=str,
        default=None,
        help="Precision of the retriever",
    )
    arg_parser.add_argument(
        "--document_index_precision",
        type=str,
        default=None,
        help="Precision of the document index",
    )
    arg_parser.add_argument(
        "--reader_precision", type=str, default=None, help="Precision of the reader"
    )
    arg_parser.add_argument(
        "--annotation_type", type=str, default="char", help="Type of annotation"
    )
    arg_parser.add_argument(
        "--workers", type=int, default=None, help="Number of workers"
    )
    arg_parser.add_argument(
        "--host", type=str, default="localhost", help="Host address"
    )
    arg_parser.add_argument("--port", type=int, default=8000, help="Port number")
    args = arg_parser.parse_args()

    main(**vars(args))
